"""
ISO base media file format / Apple Quicktime Movie parser.

Documents:
- Parsing and Writing QuickTime Files in Java (by Chris Adamson, 02/19/2003)
  http://www.onjava.com/pub/a/onjava/2003/02/19/qt_file_format.html
- QuickTime File Format (official technical reference)
  http://developer.apple.com/documentation/QuickTime/QTFF/qtff.pdf
- Apple QuickTime:
  http://wiki.multimedia.cx/index.php?title=Apple_QuickTime
- File type (ftyp):
  http://www.ftyps.com/
- MPEG4 standard
  http://neuron2.net/library/avc/c041828_ISO_IEC_14496-12_2005%28E%29.pdf

Author: Victor Stinner, Robert Xiao
Creation: 2 august 2006
"""

from hachoir.parser import Parser
from hachoir.parser.common.win32 import GUID
from hachoir.field import (ParserError, FieldSet, MissingField,
                           Enum,
                           Bit, NullBits, Bits, UInt8, Int16, UInt16, UInt24, Int32, UInt32, Int64, UInt64, TimestampMac32,
                           String, PascalString8, PascalString16, CString,
                           RawBytes, NullBytes)
from hachoir.field.timestamp import timestampFactory
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import textHandler

from hachoir.core.tools import MAC_TIMESTAMP_T0, timedelta


# ISO/IEC 14496-1:2010 8.3.3
class InstanceLength(Bits):
    def __init__(self, parent, name, description=None):
        Bits.__init__(self, parent, name, 8, description)

        stream = self._parent.stream

        addr = self.absolute_address
        size = 8
        byte = stream.readBits(addr, 8, BIG_ENDIAN)

        value = byte & 0x7F
        while byte & 0x80:
            addr += 8
            size += 8
            byte = stream.readBits(addr, 8, BIG_ENDIAN)
            value = (value << 7) + (byte & 0x7F)
        self._size = size
        self.createValue = lambda: value


# ISO/IEC 14496-1:2010 7.2.6.5
ES_DescrTag = 0x03


def ESDescriptor(self):
    yield UInt16(self, "ES_ID")
    yield Bit(self, "streamDependenceFlag")
    yield Bit(self, "URL_Flag")
    yield Bit(self, "OCRstreamFlag")
    yield Bits(self, "streamPriority", 5)
    if self["streamDependenceFlag"].value:
        yield UInt16(self, "dependsOn_ES_ID")
    if self["URL_Flag"].value:
        yield PascalString8(self, "URL")
    if self["OCRstreamFlag"].value:
        yield UInt16(self, "OCR_ES_Id")

    yield Descriptor(self, "decConfigDescr", restrict=DecoderConfigDescriptor)

    # TODO
    while not self.eof:
        yield Descriptor(self, "descr[]")


# ISO/IEC 14496-1:2010 7.2.6.6
DecoderConfigDescrTag = 0x04


def DecoderConfigDescriptor(self):
    yield UInt8(self, "objectTypeIndication")
    yield Bits(self, "streamType", 6)
    yield Bit(self, "upStream", 1)
    yield NullBits(self, "reserved", 1)
    yield UInt24(self, "bufferSizeDB")
    yield UInt32(self, "maxBitrate")
    yield UInt32(self, "avgBitrate")

    # TODO
    while not self.eof:
        yield Descriptor(self, "descr[]")


# ISO/IEC 14496-1:2010 7.2.2.2
class Descriptor(FieldSet):
    # TODO: this is very annoying to represent without backtracking
    handlers = {
        DecoderConfigDescrTag: DecoderConfigDescriptor,
        ES_DescrTag: ESDescriptor,
    }

    def __init__(self, parent, name, description=None, restrict=None):
        FieldSet.__init__(self, parent, name, description)
        self.restrict = restrict
        field = self["sizeOfInstance"]
        self._size = field.address + field.size + field.value * 8

    def createFields(self):
        yield UInt8(self, "tag")
        yield InstanceLength(self, "sizeOfInstance")

        handler = self.handlers.get(self["tag"].value)
        if self.restrict and handler != self.restrict:
            raise ParserError("invalid descriptor")

        if handler:
            yield from handler(self)
        else:
            yield RawBytes(self, "data", self["sizeOfInstance"].value)


def timestampMac64(value):
    if not isinstance(value, (float, int)):
        raise TypeError("an integer or float is required")
    return MAC_TIMESTAMP_T0 + timedelta(seconds=value)


TimestampMac64 = timestampFactory("TimestampMac64", timestampMac64, 64)


def fixedFloatFactory(name, int_bits, float_bits, doc):
    size = int_bits + float_bits

    class Float(FieldSet):
        static_size = size
        __doc__ = doc

        def createFields(self):
            yield Bits(self, "int_part", int_bits)
            yield Bits(self, "float_part", float_bits)

        def createValue(self):
            return self["int_part"].value + float(self["float_part"].value) / (1 << float_bits)
    klass = Float
    klass.__name__ = name
    return klass


QTFloat16 = fixedFloatFactory("QTFloat32", 8, 8, "8.8 fixed point number")
QTFloat32 = fixedFloatFactory("QTFloat32", 16, 16, "16.16 fixed point number")
QTFloat2_30 = fixedFloatFactory(
    "QTFloat2_30", 2, 30, "2.30 fixed point number")


class AtomList(FieldSet):

    def createFields(self):
        while not self.eof:
            yield Atom(self, "atom[]")


# ISO/IEC 14496-12:2012 8.3.2
class TrackHeader(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version (0 or 1)")
        yield NullBits(self, "flags", 20)
        yield Bit(self, "is_in_poster")
        yield Bit(self, "is_in_preview", "Is this track used when previewing the presentation?")
        yield Bit(self, "is_in_movie", "Is this track used in the presentation?")
        yield Bit(self, "is_enabled", "Is this track enabled?")

        if self['version'].value == 0:
            # 32-bit version
            yield TimestampMac32(self, "creation_date", "Creation time of this track")
            yield TimestampMac32(self, "lastmod_date", "Last modification time of this track")
            yield UInt32(self, "track_id", "Unique nonzero identifier of this track within the presentation")
            yield NullBytes(self, "reserved[]", 4)
            yield UInt32(self, "duration", "Length of track, in movie time-units")
        elif self['version'].value == 1:
            # 64-bit version
            yield TimestampMac64(self, "creation_date", "Creation time of this track")
            yield TimestampMac64(self, "lastmod_date", "Last modification time of this track")
            yield UInt32(self, "track_id", "Unique nonzero identifier of this track within the presentation")
            yield NullBytes(self, "reserved[]", 4)
            yield UInt64(self, "duration", "Length of track, in movie time-units")
        yield NullBytes(self, "reserved[]", 8)
        yield Int16(self, "video_layer", "Middle layer is 0; lower numbers are closer to the viewer")
        yield Int16(self, "alternate_group", "Group ID that this track belongs to (0=no group)")
        yield QTFloat16(self, "volume", "Track relative audio volume (1.0 = full)")
        yield NullBytes(self, "reserved[]", 2)
        yield QTFloat32(self, "geom_a", "Width scale")
        yield QTFloat32(self, "geom_b", "Width rotate")
        yield QTFloat2_30(self, "geom_u", "Width angle")
        yield QTFloat32(self, "geom_c", "Height rotate")
        yield QTFloat32(self, "geom_d", "Height scale")
        yield QTFloat2_30(self, "geom_v", "Height angle")
        yield QTFloat32(self, "geom_x", "Position X")
        yield QTFloat32(self, "geom_y", "Position Y")
        yield QTFloat2_30(self, "geom_w", "Divider scale")
        yield QTFloat32(self, "frame_size_width")
        yield QTFloat32(self, "frame_size_height")


class TrackReferenceType(FieldSet):

    def createFields(self):
        while not self.eof:
            yield UInt32(self, "track_id[]", "Referenced track ID")


class Handler(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 24)
        yield String(self, "creator", 4)
        yield String(self, "subtype", 4)
        yield String(self, "manufacturer", 4)
        yield UInt32(self, "res_flags")
        yield UInt32(self, "res_flags_mask")
        if self.root.is_mpeg4:
            yield CString(self, "name", charset="UTF-8")
        else:
            yield PascalString8(self, "name")


class LanguageCode(FieldSet):
    static_size = 16
    MAC_LANG = {
        0: 'English',
        1: 'French',
        2: 'German',
        3: 'Italian',
        4: 'Dutch',
        5: 'Swedish',
        6: 'Spanish',
        7: 'Danish',
        8: 'Portuguese',
        9: 'Norwegian',
        10: 'Hebrew',
        11: 'Japanese',
        12: 'Arabic',
        13: 'Finnish',
        14: 'Greek',
        15: 'Icelandic',
        16: 'Maltese',
        17: 'Turkish',
        18: 'Croatian',
        19: 'Traditional Chinese',
        20: 'Urdu',
        21: 'Hindi',
        22: 'Thai',
        23: 'Korean',
        24: 'Lithuanian',
        25: 'Polish',
        26: 'Hungarian',
        27: 'Estonian',
        28: 'Latvian',
        29: 'Lappish',
        30: 'Faeroese',
        31: 'Farsi',
        32: 'Russian',
        33: 'Simplified Chinese',
        34: 'Flemish',
        35: 'Irish',
        36: 'Albanian',
        37: 'Romanian',
        38: 'Czech',
        39: 'Slovak',
        40: 'Slovenian',
        41: 'Yiddish',
        42: 'Serbian',
        43: 'Macedonian',
        44: 'Bulgarian',
        45: 'Ukrainian',
        46: 'Byelorussian',
        47: 'Uzbek',
        48: 'Kazakh',
        49: 'Azerbaijani',
        50: 'AzerbaijanAr',
        51: 'Armenian',
        52: 'Georgian',
        53: 'Moldavian',
        54: 'Kirghiz',
        55: 'Tajiki',
        56: 'Turkmen',
        57: 'Mongolian',
        58: 'MongolianCyr',
        59: 'Pashto',
        60: 'Kurdish',
        61: 'Kashmiri',
        62: 'Sindhi',
        63: 'Tibetan',
        64: 'Nepali',
        65: 'Sanskrit',
        66: 'Marathi',
        67: 'Bengali',
        68: 'Assamese',
        69: 'Gujarati',
        70: 'Punjabi',
        71: 'Oriya',
        72: 'Malayalam',
        73: 'Kannada',
        74: 'Tamil',
        75: 'Telugu',
        76: 'Sinhalese',
        77: 'Burmese',
        78: 'Khmer',
        79: 'Lao',
        80: 'Vietnamese',
        81: 'Indonesian',
        82: 'Tagalog',
        83: 'MalayRoman',
        84: 'MalayArabic',
        85: 'Amharic',
        86: 'Tigrinya',
        88: 'Somali',
        89: 'Swahili',
        90: 'Ruanda',
        91: 'Rundi',
        92: 'Chewa',
        93: 'Malagasy',
        94: 'Esperanto',
        128: 'Welsh',
        129: 'Basque',
        130: 'Catalan',
        131: 'Latin',
        132: 'Quechua',
        133: 'Guarani',
        134: 'Aymara',
        135: 'Tatar',
        136: 'Uighur',
        137: 'Dzongkha',
        138: 'JavaneseRom',
    }

    def fieldHandler(self, field):
        if field.value == 0:
            return ' '
        return chr(field.value + 0x60)

    def createFields(self):
        value = self.stream.readBits(self.absolute_address, 16, self.endian)
        if value < 1024:
            yield Enum(UInt16(self, "lang"), self.MAC_LANG)
        else:
            yield NullBits(self, "padding[]", 1)
            yield textHandler(Bits(self, "lang[0]", 5), self.fieldHandler)
            yield textHandler(Bits(self, "lang[1]", 5), self.fieldHandler)
            yield textHandler(Bits(self, "lang[2]", 5), self.fieldHandler)

    def createValue(self):
        if 'lang' in self:
            return self['lang'].display
        return self['lang[0]'].display + self['lang[1]'].display + self['lang[2]'].display


class MediaHeader(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version (0 or 1)")
        yield NullBits(self, "flags", 24)
        if self['version'].value == 0:
            # 32-bit version
            yield TimestampMac32(self, "creation_date", "Creation time of this media")
            yield TimestampMac32(self, "lastmod_date", "Last modification time of this media")
            yield UInt32(self, "time_scale", "Number of time-units per second")
            yield UInt32(self, "duration", "Length of media, in time-units")
        elif self['version'].value == 1:
            # 64-bit version
            yield TimestampMac64(self, "creation_date", "Creation time of this media")
            yield TimestampMac64(self, "lastmod_date", "Last modification time of this media")
            yield UInt32(self, "time_scale", "Number of time-units per second")
            yield UInt64(self, "duration", "Length of media, in time-units")
        yield LanguageCode(self, "language")
        yield Int16(self, "quality")


class VideoMediaHeader(FieldSet):
    GRAPHICSMODE = {
        0: ('Copy', "Copy the source image over the destination"),
        0x20: ('Blend', "Blend of source and destination; blending factor is controlled by op color"),
        0x24: ('Transparent', "Replace destination pixel with source pixel if the source pixel is not the op color"),
        0x40: ('Dither copy', "Dither image if necessary, else copy"),
        0x100: ('Straight alpha', "Blend of source and destination; blending factor is controlled by alpha channel"),
        0x101: ('Premul white alpha', "Remove white from each pixel and blend"),
        0x102: ('Premul black alpha', "Remove black from each pixel and blend"),
        0x103: ('Composition', "Track drawn offscreen and dither copied onto screen"),
        0x104: ('Straight alpha blend', "Blend of source and destination; blending factor is controlled by combining alpha channel and op color")
    }

    def graphicsDisplay(self, field):
        if field.value in self.GRAPHICSMODE:
            return self.GRAPHICSMODE[field.value][0]
        return hex(field.value)

    def graphicsDescription(self, field):
        if field.value in self.GRAPHICSMODE:
            return self.GRAPHICSMODE[field.value][1]
        return ""

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield Bits(self, "flags", 24, "Flags (=1)")
        graphics = UInt16(self, "graphicsmode")
        graphics.createDisplay = lambda: self.graphicsDisplay(graphics)
        graphics.createDescription = lambda: self.graphicsDescription(graphics)
        yield graphics
        yield UInt16(self, "op_red", "Red value for graphics mode")
        yield UInt16(self, "op_green", "Green value for graphics mode")
        yield UInt16(self, "op_blue", "Blue value for graphics mode")


class SoundMediaHeader(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 24)
        yield QTFloat16(self, "balance")
        yield UInt16(self, "reserved[]")


class HintMediaHeader(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 24)
        yield UInt16(self, "max_pdu_size")
        yield UInt16(self, "avg_pdu_size")
        yield UInt32(self, "max_bit_rate")
        yield UInt32(self, "avg_bit_rate")
        yield UInt32(self, "reserved[]")


# ISO/IEC 14496-12:2012 8.7.2
class DataEntryUrl(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 23)
        yield Bit(self, "is_same_file", "Is the reference to this file?")
        if not self['is_same_file'].value:
            yield CString(self, "location")


class DataEntryUrn(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 23)
        yield Bit(self, "is_same_file", "Is the reference to this file?")
        if not self['is_same_file'].value:
            yield CString(self, "name")
            yield CString(self, "location")


class DataReference(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count")
        for i in range(self['count'].value):
            yield Atom(self, "atom[]")


# ISO/IEC 14496-12:2012 8.6.5
class EditList(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version (0 or 1)")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count")
        version = self['version'].value
        if version == 0:
            UInt, Int = UInt32, Int32
        elif version == 1:
            UInt, Int = UInt64, Int64
        else:
            raise ParserError("elst version %d not supported" % version)
        for i in range(self['count'].value):
            yield UInt(self, "duration[]", "Duration of this edit segment")
            yield Int(self, "time[]", "Starting time of this edit segment within the media (-1 = empty edit)")
            yield QTFloat32(self, "play_speed[]", "Playback rate (0 = dwell edit, 1 = normal playback)")


class Load(FieldSet):

    def createFields(self):
        yield UInt32(self, "start")
        yield UInt32(self, "length")
        # PreloadAlways = 1 or TrackEnabledPreload = 2
        yield UInt32(self, "flags")
        # KeepInBuffer = 0x00000004; HighQuality = 0x00000100; SingleFieldVideo = 0x00100000
        yield UInt32(self, "hints")


# ISO/IEC 14496-12:2012 8.2.2
class MovieHeader(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version (0 or 1)")
        yield NullBits(self, "flags", 24)
        if self['version'].value == 0:
            # 32-bit version
            yield TimestampMac32(self, "creation_date", "Creation time of this presentation")
            yield TimestampMac32(self, "lastmod_date", "Last modification time of this presentation")
            yield UInt32(self, "time_scale", "Number of time-units per second")
            yield UInt32(self, "duration", "Length of presentation, in time-units")
        elif self['version'].value == 1:
            # 64-bit version
            yield TimestampMac64(self, "creation_date", "Creation time of this presentation")
            yield TimestampMac64(self, "lastmod_date", "Last modification time of this presentation")
            yield UInt32(self, "time_scale", "Number of time-units per second")
            yield UInt64(self, "duration", "Length of presentation, in time-units")
        yield QTFloat32(self, "play_speed", "Preferred playback speed (1.0 = normal)")
        yield QTFloat16(self, "volume", "Preferred playback volume (1.0 = full)")
        yield NullBytes(self, "reserved[]", 10)
        yield QTFloat32(self, "geom_a", "Width scale")
        yield QTFloat32(self, "geom_b", "Width rotate")
        yield QTFloat2_30(self, "geom_u", "Width angle")
        yield QTFloat32(self, "geom_c", "Height rotate")
        yield QTFloat32(self, "geom_d", "Height scale")
        yield QTFloat2_30(self, "geom_v", "Height angle")
        yield QTFloat32(self, "geom_x", "Position X")
        yield QTFloat32(self, "geom_y", "Position Y")
        yield QTFloat2_30(self, "geom_w", "Divider scale")
        yield UInt32(self, "preview_start")
        yield UInt32(self, "preview_length")
        yield UInt32(self, "still_poster")
        yield UInt32(self, "sel_start")
        yield UInt32(self, "sel_length")
        yield UInt32(self, "current_time")
        yield UInt32(self, "next_track_ID", "Value to use as the track ID for the next track added")


# ISO/IEC 14496-12:2012 4.3
class FileType(FieldSet):

    def createFields(self):
        yield String(self, "brand", 4, "Major brand")
        yield UInt32(self, "version", "Version")
        while not self.eof:
            yield String(self, "compat_brand[]", 4, "Compatible brand")


# ISO/IEC 14496-12:2012 8.8.5
class MovieFragmentHeader(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "sequence_number")


# ISO/IEC 14496-12:2012 8.8.7
class TrackFragmentHeaderBox(FieldSet):
    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield UInt24(self, "flags")
        flags = self["flags"].value

        yield UInt32(self, "track_ID")
        if flags & 0x1:
            yield UInt64(self, "base_data_offset")
        if flags & 0x2:
            yield UInt32(self, "sample_description_index")
        if flags & 0x8:
            yield UInt32(self, "default_sample_duration")
        if flags & 0x10:
            yield UInt32(self, "default_sample_size")
        if flags & 0x20:
            yield UInt32(self, "default_sample_flags")


# ISO/IEC 14496-12:2012 8.8.8
class TrackRunSample(FieldSet):
    def createFields(self):
        flags = self["../flags"].value
        if flags & 0x100:
            yield UInt32(self, "sample_duration")
        if flags & 0x200:
            yield UInt32(self, "sample_size")
        if flags & 0x400:
            yield UInt32(self, "sample_flags")
        if flags & 0x800:
            yield UInt32(self, "sample_composition_time_offset")


class TrackRunBox(FieldSet):
    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield UInt24(self, "flags")
        flags = self["flags"].value

        yield UInt32(self, "sample_count")
        if flags & 0x1:
            yield UInt32(self, "data_offset")
        if flags & 0x4:
            yield UInt32(self, "first_sample_flags")
        for i in range(self["sample_count"].value):
            yield TrackRunSample(self, "sample[]")


# ISO/IEC 14496-12:2012 8.8.10
class TrackFragmentRandomAccess(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "track_id")
        yield NullBits(self, "reserved", 26)
        yield Bits(self, "length_size_of_traf_num", 2)
        yield Bits(self, "length_size_of_trun_num", 2)
        yield Bits(self, "length_size_of_sample_num", 2)
        yield UInt32(self, "number_of_entry")
        for i in range(self['number_of_entry'].value):
            if self['version'].value == 1:
                yield UInt64(self, "time[%i]" % i)
                yield UInt64(self, "moof_offset[%i]" % i)
            else:
                yield UInt32(self, "time[%i]" % i)
                yield UInt32(self, "moof_offset[%i]" % i)

            if self['length_size_of_traf_num'].value == 3:
                yield UInt64(self, "traf_number[%i]" % i)
            elif self['length_size_of_traf_num'].value == 2:
                yield UInt32(self, "traf_number[%i]" % i)
            elif self['length_size_of_traf_num'].value == 1:
                yield UInt16(self, "traf_number[%i]" % i)
            else:
                yield UInt8(self, "traf_number[%i]" % i)

            if self['length_size_of_trun_num'].value == 3:
                yield UInt64(self, "trun_number[%i]" % i)
            elif self['length_size_of_trun_num'].value == 2:
                yield UInt32(self, "trun_number[%i]" % i)
            elif self['length_size_of_trun_num'].value == 1:
                yield UInt16(self, "trun_number[%i]" % i)
            else:
                yield UInt8(self, "trun_number[%i]" % i)

            if self['length_size_of_sample_num'].value == 3:
                yield UInt64(self, "sample_number[%i]" % i)
            elif self['length_size_of_sample_num'].value == 2:
                yield UInt32(self, "sample_number[%i]" % i)
            elif self['length_size_of_sample_num'].value == 1:
                yield UInt16(self, "sample_number[%i]" % i)
            else:
                yield UInt8(self, "sample_number[%i]" % i)


# ISO/IEC 14496-12:2012 8.8.11
class MovieFragmentRandomAccessOffset(FieldSet):

    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "size")


def findHandler(self):
    ''' find the handler corresponding to this fieldset '''
    while self:
        if self.name in ('media', 'tags'):
            break
        self = self.parent
    else:
        return None
    for atom in self:
        if atom['tag'].value == 'hdlr':
            return atom['hdlr']
    return None


class METATAG(FieldSet):

    def createFields(self):
        yield UInt8(self, "unk[]", "0x80 or 0x00")
        yield PascalString16(self, "tag_name", charset='UTF-8')
        yield UInt16(self, "unk[]", "0x0001")
        yield UInt16(self, "unk[]", "0x0000")
        yield PascalString16(self, "tag_value", charset='UTF-8')


class META(FieldSet):

    def createFields(self):
        # This tag has too many variant forms.
        if '/tags/' in self.path:
            yield UInt32(self, "count")
            for i in range(self['count'].value):
                yield METATAG(self, "tag[]")
        elif self.stream.readBits(self.absolute_address, 32, self.endian) == 0:
            yield UInt8(self, "version")
            yield Bits(self, "flags", 24)
            yield AtomList(self, "tags")
        else:
            yield AtomList(self, "tags")


class Item(FieldSet):

    def createFields(self):
        yield UInt32(self, "size")
        yield UInt32(self, "index")
        yield Atom(self, "value")


class KeyList(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count")
        for i in range(self['count'].value):
            yield Atom(self, "key[]")


class ItemList(FieldSet):

    def createFields(self):
        handler = findHandler(self)
        if handler is None:
            raise ParserError("ilst couldn't find metadata handler")
        if handler['subtype'].value == 'mdir':
            while not self.eof:
                yield Atom(self, "atom[]")
        elif handler['subtype'].value == 'mdta':
            while not self.eof:
                yield Item(self, "item[]")


class NeroChapters(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "unknown")
        yield UInt8(self, "count", description="Number of chapters")
        for i in range(self['count'].value):
            yield UInt64(self, "chapter_start[]")
            yield PascalString8(self, "chapter_name[]", charset='UTF-8')


# ISO/IEC 14496-12:2012 8.6.1.2
class SampleDecodeTimeTable(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count", description="Total entries in sample time table")
        for i in range(self['count'].value):
            yield UInt32(self, "sample_count[]", "Number of consecutive samples with this delta")
            yield UInt32(self, "sample_delta[]", "Decode time delta since last sample, in time-units")


class SampleCompositionTimeTable(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count", description="Total entries in sample time table")
        for i in range(self['count'].value):
            yield UInt32(self, "sample_count[]", "Number of consecutive samples with this offset")
            yield UInt32(self, "sample_offset[]", "Difference between decode time and composition time of this sample, in time-units")


class ChunkOffsetTable(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count", description="Total entries in offset table")
        for i in range(self['count'].value):
            yield UInt32(self, "chunk_offset[]")


class ChunkOffsetTable64(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count", description="Total entries in offset table")
        for i in range(self['count'].value):
            yield UInt64(self, "chunk_offset[]")


# ISO/IEC 14496-14:2003 5.6
class ESDBox(FieldSet):
    def createFields(self):
        yield UInt8(self, "version", "Version")
        yield NullBits(self, "flags", 24)

        yield Descriptor(self, "ES", restrict=ESDescriptor)


# ETSI TS 102 366 v1.2.1 F.6
class EC3SpecificBoxSubstream(FieldSet):
    def createFields(self):
        yield Bits(self, "fscod", 2)
        yield Bits(self, "bsid", 5)
        yield Bits(self, "bsmod", 5)
        yield Bits(self, "acmod", 3)
        yield Bits(self, "lfeon", 1)
        yield NullBits(self, "reserved", 3)
        yield Bits(self, "num_dep_sub", 4)
        if self["num_dep_sub"].value:
            yield Bits(self, "chan_loc", 9)
        else:
            yield NullBits(self, "reserved2", 1)


class EC3SpecificBox(FieldSet):
    def createFields(self):
        yield Bits(self, "data_rate", 13)
        yield Bits(self, "num_ind_sub", 3)
        for i in range(self["num_ind_sub"].value + 1):
            yield EC3SpecificBoxSubstream(self, "sub[]")


# ISO/IEC 14496-15:2014 5.3.3.1
class AVCDecoderConfigurationRecord(FieldSet):
    def createFields(self):
        yield UInt8(self, "configurationVersion")
        yield UInt8(self, "AVCProfileIndication")
        yield UInt8(self, "profile_compatibility")
        yield UInt8(self, "AVCLevelIndication")
        yield NullBits(self, "reserved[]", 6)
        yield Bits(self, "lengthSizeMinusOne", 2)
        yield NullBits(self, "reserved[]", 3)

        yield Bits(self, "numOfSequenceParameterSets", 5)
        for i in range(self["numOfSequenceParameterSets"].value):
            yield PascalString16(self, "sequenceParameterSetNALUnit[]")

        yield UInt8(self, "numOfPictureParameterSets")
        for i in range(self["numOfPictureParameterSets"].value):
            yield PascalString16(self, "pictureParameterSetNALUnit[]")

        if self['AVCProfileIndication'].value in (100, 110, 122, 144) and not self.eof:
            yield NullBits(self, "reserved[]", 6)
            yield Bits(self, "chroma_format", 2)
            yield NullBits(self, "reserved[]", 5)
            yield Bits(self, "bit_depth_luma_minus8", 3)
            yield NullBits(self, "reserved[]", 5)
            yield Bits(self, "bit_depth_chroma_minus8", 3)

            yield UInt8(self, "numOfSequenceParameterSetExt")
            for i in range(self["numOfSequenceParameterSetExt"].value):
                yield PascalString16(self, "sequenceParameterSetExtNALUnit[]")


# ISO/IEC 14496-15:2014 5.4.2.1
class MPEG4BitRateBox(FieldSet):
    def createFields(self):
        yield UInt32(self, "bufferSizeDB")
        yield UInt32(self, "maxBitrate")
        yield UInt32(self, "avgBitrate")


class AVCConfigurationBox(FieldSet):
    def createFields(self):
        yield AVCDecoderConfigurationRecord(self, "AVCConfig")


# ISO/IEC 14496-12:2012 8.5.2.2
def VisualSampleEntry(self):
    yield UInt16(self, "version")
    yield UInt16(self, "revision_level")
    yield RawBytes(self, "vendor_id", 4)
    yield UInt32(self, "temporal_quality")
    yield UInt32(self, "spatial_quality")
    yield UInt16(self, "width", "Width (pixels)")
    yield UInt16(self, "height", "Height (pixels)")
    yield QTFloat32(self, "horizontal_resolution", "Horizontal resolution in DPI")
    yield QTFloat32(self, "vertical resolution", "Vertical resolution in DPI")
    yield UInt32(self, "data_size")
    yield UInt16(self, "frame_count")
    yield UInt8(self, "compressor_name_length")
    yield String(self, "compressor_name", 31, strip='\0')
    yield UInt16(self, "depth", "Bit depth of image")
    yield Int16(self, "unknown")


def AudioSampleEntry(self):
    yield NullBytes(self, "reserved[]", 8)
    yield UInt16(self, "channels", "Number of audio channels")
    yield UInt16(self, "samplesize", "Sample size in bits")
    yield UInt16(self, "unknown")
    yield NullBytes(self, "reserved[]", 2)
    yield QTFloat32(self, "samplerate", "Sample rate in Hz")


class SampleEntry(FieldSet):

    def createFields(self):
        yield UInt32(self, "size")
        yield RawBytes(self, "format", 4, "Data Format (codec)")
        yield NullBytes(self, "reserved[]", 6, "Reserved")
        yield UInt16(self, "data_reference_index")
        handler = findHandler(self)
        if not handler:
            raise ParserError("stsd couldn't find track handler")
        if handler['subtype'].value == 'soun':
            yield from AudioSampleEntry(self)
        elif handler['subtype'].value == 'vide':
            yield from VisualSampleEntry(self)
        elif handler['subtype'].value == 'hint':
            # Hint sample entry
            pass

        if self["format"].value in (b"enca", b"encv", b"mp4a", b"mp4v", b"mp4s", b"avc1"):
            # MP4VisualSampleEntry, MP4AudioSampleEntry, MpegSampleEntry, EC3SampleEntry, AVCSampleEntry...
            # all just have appended atoms
            while not self.eof:
                yield Atom(self, "atom[]")
        else:
            size = self['size'].value - self.current_size // 8
            if size > 0:
                yield RawBytes(self, "extra_data", size)


# ISO/IEC 14496-12:2012 8.5.2
class SampleDescription(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count", description="Total entries in table")
        for i in range(self['count'].value):
            yield SampleEntry(self, "sample_entry[]")


# ISO/IEC 14496-12:2012 8.6.2
class SyncSampleTable(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count", description="Number of sync samples")
        for i in range(self['count'].value):
            yield UInt32(self, "sample_number[]")


class SampleSizeTable(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "uniform_size", description="Uniform size of each sample (0 if non-uniform)")
        yield UInt32(self, "count", description="Number of samples")
        if self['uniform_size'].value == 0:
            for i in range(self['count'].value):
                yield UInt32(self, "sample_size[]")


class CompactSampleSizeTable(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield NullBits(self, "reserved[]", 24)
        yield UInt8(self, "field_size", "Size of each entry in this table, in bits")
        yield UInt32(self, "count", description="Number of samples")
        bitsize = self['field_size'].value
        for i in range(self['count'].value):
            yield Bits(self, "sample_size[]", bitsize)
        if self.current_size % 8 != 0:
            yield NullBits(self, "padding[]", 8 - (self.current_size % 8))


# ISO/IEC 14496-12:2012 8.7.4
class SampleToChunkTable(FieldSet):

    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "count", description="Number of samples")
        for i in range(self['count'].value):
            yield UInt32(self, "first_chunk[]")
            yield UInt32(self, "samples_per_chunk[]")
            yield UInt32(self, "sample_description_index[]")


# ISO/IEC 14496-12:2012 8.12.1
class ProtectionSchemeInfoBox(FieldSet):
    def createFields(self):
        yield Atom(self, "original_format")
        if not self.eof:
            yield Atom(self, "scheme_type_box")
        if not self.eof:
            yield Atom(self, "info")


# ISO/IEC 14496-12:2012 8.12.2
class OriginalFormatBox(FieldSet):
    def createFields(self):
        yield RawBytes(self, "data_format", 4)


# ISO/IEC 14496-12:2012 8.12.5
class SchemeTypeBox(FieldSet):
    def createFields(self):
        yield UInt8(self, "version")
        yield UInt24(self, "flags")
        yield RawBytes(self, "scheme_type", 4)
        yield UInt32(self, "scheme_version")
        if self["flags"].value & 0x1:
            yield CString(self, "scheme_uri")


# ISO/IEC 14496-12:2012 8.12.6
class SchemeInformationBox(FieldSet):
    def createFields(self):
        yield Atom(self, "scheme_specific_data")


# ISO/IEC 14496-12:2012 8.16.3
class SegmentIndexBoxReference(FieldSet):
    def createFields(self):
        yield Bit(self, "reference_type")
        yield Bits(self, "referenced_size", 31)
        yield UInt32(self, "subsegment_duration")
        yield Bit(self, "starts_with_SAP")
        yield Bits(self, "SAP_type", 3)
        yield Bits(self, "SAP_delta_time", 28)


class SegmentIndexBox(FieldSet):
    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield UInt32(self, "reference_ID")
        yield UInt32(self, "timescale")
        if self["version"].value == 0:
            yield UInt32(self, "earliest_presentation_time")
            yield UInt32(self, "first_offset")
        else:
            yield UInt64(self, "earliest_presentation_time")
            yield UInt64(self, "first_offset")
        yield NullBits(self, "reserved", 16)
        yield UInt16(self, "reference_count")
        for i in range(self["reference_count"].value):
            yield SegmentIndexBoxReference(self, "reference[]")


# ISO/IEC 23001-7:2016 7.2
class SampleEncryptionItem(FieldSet):
    def createFields(self):
        yield RawBytes(self, "IV", 8)  # TODO: Per_Sample_IV_Size
        if self["../flags"].value & 0x2:
            yield UInt16(self, "subsample_count")
            for i in range(self["subsample_count"].value):
                yield UInt16(self, "BytesOfClearData[]")
                yield UInt32(self, "ByteOfProtectedData[]")


class SampleEncryptionBox(FieldSet):
    def createFields(self):
        yield UInt8(self, "version")
        yield UInt24(self, "flags")

        yield UInt32(self, "sample_count")
        for i in range(self["sample_count"].value):
            yield SampleEncryptionItem(self, "sample[]")


# ISO/IEC 23001-7:2016 8.1.1
class ProtectionSystemSpecificHeaderBox(FieldSet):
    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)

        yield RawBytes(self, "SystemID", 16)

        if self["version"].value > 0:
            yield UInt32(self, "KID_Count")
            for i in range(self["KID_Count"].value):
                yield RawBytes(self, "KID[]", 16)

        yield UInt32(self, "DataSize")
        yield RawBytes(self, "Data", self["DataSize"].value)


# ISO/IEC 23001-7:2016 8.2
class TrackEncryptionBox(FieldSet):
    def createFields(self):
        yield UInt8(self, "version")
        yield NullBits(self, "flags", 24)
        yield NullBits(self, "reserved", 8)
        if self["version"].value == 0:
            yield NullBits(self, "reserved2", 8)
        else:
            yield Bits(self, "default_crypt_byte_block", 4)
            yield Bits(self, "default_skip_byte_block", 4)
        yield UInt8(self, "default_isProtected")
        yield UInt8(self, "default_Per_sample_IV_Size")
        yield RawBytes(self, "default_KID", 16)
        if self["default_isProtected"].value == 1 and self["default_Per_sample_IV_Size"].value == 0:
            yield UInt8(self, "default_constant_IV_size")
            yield RawBytes(self, "default_constant_IV", self["default_constant_IV_size"].value)


class Atom(FieldSet):
    tag_info = {
        "ftyp": (FileType, "file_type", "File type and compatibility"),
        # pdin: progressive download information
        # pnot: movie preview (old QT spec)
        "moov": (AtomList, "movie", "Container for all metadata"),
            "mvhd": (MovieHeader, "movie_hdr", "Movie header, overall declarations"),
            # clip: movie clipping (old QT spec)
                # crgn: movie clipping region (old QT spec)
            "trak": (AtomList, "track", "Container for an individual track or stream"),
                "tkhd": (TrackHeader, "track_hdr", "Track header, overall information about the track"),
                # matt: track matte (old QT spec)
                    # kmat: compressed matte (old QT spec)
                "tref": (AtomList, "tref", "Track reference container"),
                    "hint": (TrackReferenceType, "hint", "Original media track(s) for this hint track"),
                    "cdsc": (TrackReferenceType, "cdsc", "Reference to track described by this track"),
                "edts": (AtomList, "edts", "Edit list container"),
                    "elst": (EditList, "elst", "Edit list"),
                "load": (Load, "load", "Track loading settings (old QT spec)"),
                # imap: Track input map (old QT spec)
                "senc": (SampleEncryptionBox, "senc", "Sample encryption information"),
                "mdia": (AtomList, "media", "Container for the media information in a track"),
                    "mdhd": (MediaHeader, "media_hdr", "Media header, overall information about the media"),
                    "hdlr": (Handler, "hdlr", "Handler, declares the media or metadata (handler) type"),
                    "minf": (AtomList, "minf", "Media information container"),
                        "vmhd": (VideoMediaHeader, "vmhd", "Video media header, overall information (video track only)"),
                        "smhd": (SoundMediaHeader, "smhd", "Sound media header, overall information (sound track only)"),
                        "hmhd": (HintMediaHeader, "hmhd", "Hint media header, overall information (hint track only)"),
                        # nmhd: Null media header, overall information (some tracks only) (unparsed)
                        "dinf": (AtomList, "dinf", "Data information, container"),
                            "dref": (DataReference, "dref", "Data reference, declares source(s) of media data in track"),
                                "url ": (DataEntryUrl, "url", "URL data reference"),
                                "urn ": (DataEntryUrn, "urn", "URN data reference"),
                        "stbl": (AtomList, "stbl", "Sample table, container for the time/space map"),
                            "stsd": (SampleDescription, "stsd", "Sample descriptions (codec types, initialization etc.)"),
                                "esds": (ESDBox, "esds", "Elementary stream descriptor"),
                                "avcC": (AVCConfigurationBox, "avcC", "AVC configuration"),
                                "btrt": (MPEG4BitRateBox, "btrt", "AVC stream bitrate"),
                                "dec3": (EC3SpecificBox, "dec3", "Enhanced AC-3 speicifc information"),
                                "sinf": (ProtectionSchemeInfoBox, "sinf", "Protection scheme information"),
                                    "frma": (OriginalFormatBox, "frma", "original format"),
                                    "schm": (SchemeTypeBox, "schm", "scheme type"),
                                    "schi": (SchemeInformationBox, "schi", "scheme information"),
                                        "tenc": (TrackEncryptionBox, "tenc", "track encryption"),
                            "stts": (SampleDecodeTimeTable, "stts", "decoding time-to-sample delta table"),
                            "ctts": (SampleCompositionTimeTable, "ctts", "composition time-to-sample offset table"),
                            "stsc": (SampleToChunkTable, "stsc", "sample-to-chunk, partial data-offset information"),
                            "stsz": (SampleSizeTable, "stsz", "Sample size table (framing)"),
                            "stz2": (CompactSampleSizeTable, "stz2", "Compact sample size table (framing)"),
                            "stco": (ChunkOffsetTable, "stco", "Chunk offset, partial data-offset information"),
                            "co64": (ChunkOffsetTable64, "co64", "64-bit chunk offset"),
                            "stss": (SyncSampleTable, "stss", "Sync sample table (random access points)"),
                            # stsh: shadow sync sample table
                            # padb: sample padding bits
                            # stdp: sample degradation priority
                            # sdtp: independent and disposable samples
                            # sbgp: sample-to-group
                            # sgpd: sample group description
                            # subs: sub-sample information
            # ctab color table (old QT spec)
            # mvex: movie extends
                # mehd: movie extends header
                # trex: track extends defaults
            # ipmc: IPMP control
            "pssh": (ProtectionSystemSpecificHeaderBox, "pssh", "Protection system information"),
        "moof": (AtomList, "moof", "movie fragment"),
            "mfhd": (MovieFragmentHeader, "mfhd", "movie fragment header"),
            "traf": (AtomList, "traf", "track fragment"),
                "tfhd": (TrackFragmentHeaderBox, "tfgd", "track fragment header"),
                "trun": (TrackRunBox, "trun", "track fragment run"),
                # sdtp: independent and disposable samples
                # sbgp: sample-to-group
                # subs: sub-sample information
        "mfra": (AtomList, "mfra", "movie fragment random access"),
            "tfra": (TrackFragmentRandomAccess, "tfra", "track fragment random access"),
            "mfro": (MovieFragmentRandomAccessOffset, "mfro", "movie fragment random access offset"),
        "sidx": (SegmentIndexBox, "sidx", "segment index"),
        # mdat: media data container
        # free: free space (unparsed)
        # skip: free space (unparsed)
        "udta": (AtomList, "udta", "User data"),
        "meta": (META, "meta", "File metadata"),
            "keys": (KeyList, "keys", "Metadata keys"),
            # hdlr
            # dinf
                # dref: data reference, declares source(s) of metadata items
            # ipmc: IPMP control
            # iloc: item location
            # ipro: item protection
                # sinf: protection scheme information
                    # frma: original format
                    # imif: IPMP information
                    # schm: scheme type
                    # schi: scheme information
            # iinf: item information
            # xml : XML container
            # bxml: binary XML container
            # pitm: primary item reference
        # other tags
        "ilst": (ItemList, "ilst", "Item list"),
            "trkn": (AtomList, "trkn", "Metadata: Track number"),
            "disk": (AtomList, "disk", "Metadata: Disk number"),
            "tmpo": (AtomList, "tempo", "Metadata: Tempo"),
            "cpil": (AtomList, "cpil", "Metadata: Compilation"),
            "gnre": (AtomList, "gnre", "Metadata: Genre"),
            "\xa9cpy": (AtomList, "copyright", "Metadata: Copyright statement"),
            "\xa9day": (AtomList, "date", "Metadata: Date of content creation"),
            "\xa9dir": (AtomList, "director", "Metadata: Movie director"),
            "\xa9ed1": (AtomList, "edit1", "Metadata: Edit date and description (1)"),
            "\xa9ed2": (AtomList, "edit2", "Metadata: Edit date and description (2)"),
            "\xa9ed3": (AtomList, "edit3", "Metadata: Edit date and description (3)"),
            "\xa9ed4": (AtomList, "edit4", "Metadata: Edit date and description (4)"),
            "\xa9ed5": (AtomList, "edit5", "Metadata: Edit date and description (5)"),
            "\xa9ed6": (AtomList, "edit6", "Metadata: Edit date and description (6)"),
            "\xa9ed7": (AtomList, "edit7", "Metadata: Edit date and description (7)"),
            "\xa9ed8": (AtomList, "edit8", "Metadata: Edit date and description (8)"),
            "\xa9ed9": (AtomList, "edit9", "Metadata: Edit date and description (9)"),
            "\xa9fmt": (AtomList, "format", "Metadata: Movie format (CGI, digitized, etc.)"),
            "\xa9inf": (AtomList, "info", "Metadata: Information about the movie"),
            "\xa9prd": (AtomList, "producer", "Metadata: Movie producer"),
            "\xa9prf": (AtomList, "performers", "Metadata: Performer names"),
            "\xa9req": (AtomList, "requirements", "Metadata: Special hardware and software requirements"),
            "\xa9src": (AtomList, "source", "Metadata: Credits for those who provided movie source content"),
            "\xa9nam": (AtomList, "name", "Metadata: Name of song or video"),
            "\xa9des": (AtomList, "description", "Metadata: File description"),
            "\xa9cmt": (AtomList, "comment", "Metadata: General comment"),
            "\xa9alb": (AtomList, "album", "Metadata: Album name"),
            "\xa9gen": (AtomList, "genre", "Metadata: Custom genre"),
            "\xa9ART": (AtomList, "artist", "Metadata: Artist name"),
            "\xa9too": (AtomList, "encoder", "Metadata: Encoder"),
            "\xa9wrt": (AtomList, "writer", "Metadata: Writer"),
            "covr": (AtomList, "cover", "Metadata: Cover art"),
            "----": (AtomList, "misc", "Metadata: Miscellaneous"),
        "tags": (AtomList, "tags", "File tags"),
        "tseg": (AtomList, "tseg", "tseg"),
        "chpl": (NeroChapters, "chpl", "Nero chapter data"),
    }  # noqa
    tag_handler = [item[0] for item in tag_info]
    tag_desc = [item[1] for item in tag_info]

    def createFields(self):
        yield UInt32(self, "size")
        yield String(self, "tag", 4, charset="ASCII")
        size = self["size"].value
        if size == 1:
            # 64-bit size
            yield UInt64(self, "size64")
            size = self["size64"].value - 16
        elif size == 0:
            # Unbounded atom
            if self._size is None:
                size = (self.parent.size - self.parent.current_size) // 8 - 8
            else:
                size = (self.size - self.current_size) // 8
        else:
            size = size - 8
        if self['tag'].value == 'uuid':
            yield GUID(self, "usertag")
            tag = self["usertag"].value
            size -= 16
        else:
            tag = self["tag"].value
        if size > 0:
            if tag in self.tag_info:
                handler, name, desc = self.tag_info[tag]
                yield handler(self, name, desc, size=size * 8)
            else:
                yield RawBytes(self, "data", size)

    def createDescription(self):
        if self["tag"].value == "uuid":
            return "Atom: uuid: " + self["usertag"].value
        return "Atom: %s" % self["tag"].value


class MP4File(Parser):
    PARSER_TAGS = {
        "id": "mov",
        "category": "video",
        "file_ext": ("mov", "qt", "mp4", "m4v", "m4a", "m4p", "m4b"),
        "mime": ("video/quicktime", 'video/mp4'),
        "min_size": 8 * 8,
        "magic": ((b"moov", 4 * 8),),
        "description": "Apple QuickTime movie"
    }
    BRANDS = {
        # File type brand => MIME type
        'mp41': 'video/mp4',
        'mp42': 'video/mp4',
        'avc1': 'video/mp4',
        'isom': 'video/mp4',
        'iso2': 'video/mp4',
    }
    endian = BIG_ENDIAN

    def __init__(self, *args, **kw):
        Parser.__init__(self, *args, **kw)

    is_mpeg4 = property(lambda self: self.mime_type == 'video/mp4')

    def validate(self):
        # TODO: Write better code, erk!
        size = self.stream.readBits(0, 32, self.endian)
        if size < 8:
            return "Invalid first atom size"
        tag = self.stream.readBytes(4 * 8, 4)
        if tag not in (b"ftyp", b"moov", b"free"):
            return "Unknown MOV file type"
        return True

    def createFields(self):
        while not self.eof:
            yield Atom(self, "atom[]")

    def createMimeType(self):
        first = self[0]
        try:
            # Read brands in the file type
            if first['tag'].value != "ftyp":
                return None
            file_type = first["file_type"]
            brand = file_type["brand"].value
            if brand in self.BRANDS:
                return self.BRANDS[brand]
            for field in file_type.array("compat_brand"):
                brand = field.value
                if brand in self.BRANDS:
                    return self.BRANDS[brand]
        except MissingField:
            pass
        return 'video/quicktime'



"""
Mac OS X .DS_Store parser.

Documents:
- http://search.cpan.org/~wiml/Mac-Finder-DSStore-0.95/DSStoreFormat.pod
Author: Robert Xiao
Created: 2010-09-01
"""

from hachoir.parser import HachoirParser
from hachoir.field import (RootSeekableFieldSet, FieldSet,
                           NullBytes, RawBytes, PaddingBytes, Bytes, SubFile, String, PascalString8,
                           Bits, UInt8, UInt16, UInt32,
                           Link,
                           ParserError)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import displayHandler
from hachoir.core.tools import paddingSize


class BlockAddress(FieldSet):
    static_size = 32

    def createFields(self):
        yield displayHandler(Bits(self, "offset", 27, description="Offset into file divided by 32"), lambda x: hex(x * 32).strip('L'))
        yield displayHandler(Bits(self, "size", 5, description="Power-of-2 size of the block"), lambda x: hex(1 << x).strip('L'))

    def createValue(self):
        return (self['offset'].value * 32, 1 << self['size'].value)

    def createDisplay(self):
        if self['size'].value == 0:
            return 'null block'
        return self['offset'].display + '/' + self['size'].display


class BudHeader(FieldSet):
    static_size = 32 * 8

    def createFields(self):
        yield Bytes(self, "magic", 4, description="Always Bud1")
        yield UInt32(self, "allocator_offset")
        yield UInt32(self, "allocator_size")
        yield UInt32(self, "allocator_offset_backup", description="Finder will refuse to read the file if this does not match the first copy")
        for i in range(4):
            yield BlockAddress(self, "block_address_copy[]", description="Copies of block addresses defined in the allocator")


class BudDirectory(FieldSet):

    def createFields(self):
        yield PascalString8(self, "name", charset="MacRoman")
        yield UInt32(self, "block")

    def createValue(self):
        return (self['name'].value, self['block'].value)

    def createDisplay(self):
        return self['name'].display


class FreeList(FieldSet):

    def createFields(self):
        yield UInt32(self, "count")
        for i in range(self['count'].value):
            yield UInt32(self, "offset[]")


class BudAllocator(FieldSet):

    def createFields(self):
        yield UInt32(self, "nblocks")
        yield UInt32(self, "unknown", description="Always 0")
        for i in range(self['nblocks'].value):
            yield BlockAddress(self, "block[]")
        padding = paddingSize(self['nblocks'].value, 256)
        if padding:
            yield NullBytes(self, "padding", padding * 4, description="padding to make the number of blocks a multiple of 256")
        yield UInt32(self, "ndirs")
        for i in range(self['ndirs'].value):
            yield BudDirectory(self, "dir[]")
        for i in range(32):
            yield FreeList(self, "freelist[]")
        if self.current_size < self.size:
            yield PaddingBytes(self, "slack", (self.size - self.current_size) // 8, description="slack space")


class DSDB(FieldSet):

    def createFields(self):
        yield UInt32(self, "root_block", "The block number of the root node of the B-tree")
        yield UInt32(self, "tree_levels", "The number of levels of internal nodes (tree height minus one)")
        yield UInt32(self, "num_records", "The number of records in the tree")
        yield UInt32(self, "num_nodes", "The number of nodes in the tree (tree nodes, not including this header block)")
        yield UInt32(self, "unknown", "Always 0x1000, probably the tree node page size")
        if self.current_size < self.size:
            yield PaddingBytes(self, "slack", (self.size - self.current_size) // 8, description="slack space")


class PascalString32UTF16(FieldSet):

    def createFields(self):
        yield UInt32(self, "size")
        yield String(self, "value", self['size'].value * 2, charset="UTF-16-BE")

    def createValue(self):
        return self['value'].value

    def createDisplay(self):
        return str(self['value'].display)


class DSRecord(FieldSet):

    def createFields(self):
        yield PascalString32UTF16(self, "filename")
        yield Bytes(self, "property", 4)
        yield Bytes(self, "type", 4)
        type = self['type'].value
        if type == 'long':
            yield UInt32(self, "value")
        elif type == 'shor':
            yield NullBytes(self, "padding", 2)
            yield UInt16(self, "value")
        elif type == 'bool':
            yield UInt8(self, "value")
        elif type == 'blob':
            yield UInt32(self, "size")
            yield SubFile(self, "value", self['size'].value)
        elif type == 'type':
            yield Bytes(self, "value", 4)
        elif type == 'ustr':
            yield PascalString32UTF16(self, "value")
        else:
            raise ParserError("Unknown record type %s" % type)

    def createValue(self):
        return (self['filename'].value, self['property'].value, self['value'].value)

    def createDisplay(self):
        return self['filename'].display + ':' + self['property'].value


class BTNode(FieldSet):

    def linkValue(self, block):
        return (lambda: self['/node[%d]' % block.value])

    def createFields(self):
        yield UInt32(self, "last_block")
        yield UInt32(self, "count")
        if self['last_block'].value != 0:
            for i in range(self['count'].value):
                block = UInt32(self, "child_block[]")
                yield block
                link = Link(self, "child_link[]")
                link.createValue = self.linkValue(block)
                yield link
                yield DSRecord(self, "record[]")
            link = Link(self, "child_link[]")
            link.createValue = self.linkValue(self['last_block'])
            yield link
        else:
            for i in range(self['count'].value):
                yield DSRecord(self, "record[]")
        if self.current_size < self.size:
            yield PaddingBytes(self, "slack", (self.size - self.current_size) // 8, description="slack space")


class DSStore(HachoirParser, RootSeekableFieldSet):
    endian = BIG_ENDIAN
    MAGIC = b'\0\0\0\1Bud1'
    PARSER_TAGS = {
        "id": "dsstore",
        "category": "misc",
        "file_ext": ("DS_Store",),
        "magic": ((MAGIC, 0),),
        "min_size": 4 + 32,  # \0\0\0\1 + 32-byte header
        "description": "Mac OS X DS_Store",
    }

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes(0, len(self.MAGIC)) != self.MAGIC:
            return "Invalid magic"
        return True

    def getBlock(self, block_number):
        return self['allocator'].array('block')[block_number].value

    def createFields(self):
        yield UInt32(self, "unknown", description="Always 1")
        yield BudHeader(self, "header")
        self.seekByte(self['header/allocator_offset'].value + 4)
        yield BudAllocator(self, "allocator", size=self['header/allocator_size'].value * 8)
        for dir in self['allocator'].array('dir'):
            if dir['name'].value == 'DSDB':
                break
        else:
            raise ParserError("DSDB not found.")
        offs, size = self.getBlock(dir['block'].value)
        self.seekByte(offs + 4)
        yield DSDB(self, "dsdb", size=size * 8)

        blocks = [self['dsdb/root_block'].value]
        while blocks:
            block = blocks.pop()
            offs, size = self.getBlock(block)
            self.seekByte(offs + 4)
            node = BTNode(self, "node[%d]" % block, size=size * 8)
            yield node
            if node['last_block'].value != 0:
                new_blocks = []
                for block in node.array('child_block'):
                    new_blocks.append(block.value)
                new_blocks.append(node['last_block'].value)
                blocks.extend(reversed(new_blocks))  # dfs
                # blocks = new_blocks[::-1] + blocks # bfs

        for i, fl in enumerate(self['allocator'].array('freelist')):
            if fl['count'].value == 0:
                continue
            for offs in fl.array('offset'):
                size = min(1 << i, self.size // 8 - offs.value - 4)
                if size > 0:
                    self.seekByte(offs.value + 4)
                    yield RawBytes(self, "free[]", size)



from time import time

DATARATE_UPDATE = 1.0   # Time slice (in second) for datarate computation


class DataRate:
    """
    Compute average speed in bits per second of a function.
    Store self.size data rates to compute good average speed.
    Don't compute average before self.min_size values are computed.
    """

    def __init__(self, offset, size=20, min_size=3):
        self.last_offset = offset
        self.last_time = time()
        self.datarates = []
        # Average bit rate
        self.average = None
        # Number of stored value used to compute average data rate
        self.size = size
        self.min_size = min_size

    def update(self, offset):
        # Compute time delta
        difftime = time() - self.last_time
        if difftime < DATARATE_UPDATE:
            # Only update each second
            return
        self.last_time = time()

        # Compute data rate
        rate = float(offset - self.last_offset) / difftime
        self.last_offset = offset

        # Update statistics
        self.datarates.append(rate)
        self.datarates = self.datarates[-self.size:]
        if self.min_size <= len(self.datarates):
            self.average = sum(self.datarates) / len(self.datarates)



"""
Utilities to convert integers and binary strings to binary (number), binary
string, number, hexadecimal, etc.
"""

from hachoir.core.endian import BIG_ENDIAN, LITTLE_ENDIAN, MIDDLE_ENDIAN
from struct import calcsize, unpack, error as struct_error


def swap16(value):
    """
    Swap byte between big and little endian of a 16 bits integer.

    >>> "%x" % swap16(0x1234)
    '3412'
    """
    return (value & 0xFF) << 8 | (value >> 8)


def swap32(value):
    """
    Swap byte between big and little endian of a 32 bits integer.

    >>> "%x" % swap32(0x12345678)
    '78563412'
    """
    value = int(value)
    return (((value & 0x000000FF) << 24) |
            ((value & 0x0000FF00) << 8) |
            ((value & 0x00FF0000) >> 8) |
            ((value & 0xFF000000) >> 24))


def arrswapmid(data):
    r"""
    Convert an array of characters from middle-endian to big-endian and
    vice-versa.

    >>> bytes(arrswapmid(b"badcfehg"))
    b'abcdefgh'
    """
    assert len(data) % 2 == 0
    ret = [b''] * len(data)
    ret[1::2] = data[0::2]
    ret[0::2] = data[1::2]
    return ret


def strswapmid(data):
    r"""
    Convert raw data from middle-endian to big-endian and vice-versa.

    >>> strswapmid(b"badcfehg")
    b'abcdefgh'
    """
    return bytes(arrswapmid(data))


def bin2long(text, endian):
    """
    Convert binary number written in a string into an integer.
    Skip characters differents than "0" and "1".

    >>> bin2long("110", BIG_ENDIAN)
    6
    >>> bin2long("110", LITTLE_ENDIAN)
    3
    >>> bin2long("11 00", LITTLE_ENDIAN)
    3
    """
    assert endian in (LITTLE_ENDIAN, BIG_ENDIAN)
    bits = [(ord(character) - ord("0"))
            for character in text if character in "01"]
    if endian is not BIG_ENDIAN:
        bits = bits[::-1]
    size = len(bits)
    assert 0 < size
    value = 0
    for bit in bits:
        value *= 2
        value += bit
    return value


def str2hex(value, prefix="", glue="", format="%02X"):
    r"""
    Convert binary string in hexadecimal (base 16).

    >>> str2hex(b"ABC")
    '414243'
    >>> str2hex(b"\xF0\xAF", glue=" ")
    'F0 AF'
    >>> str2hex(b"ABC", prefix="0x")
    '0x414243'
    >>> str2hex(b"ABC", format=r"\x%02X")
    '\\x41\\x42\\x43'
    """
    if isinstance(glue, str):
        glue = str(glue)
    if 0 < len(prefix):
        text = [prefix]
    else:
        text = []
    for character in value:
        text.append(format % character)
    return glue.join(text)


def countBits(value):
    """
    Count number of bits needed to store a (positive) integer number.

    >>> countBits(0)
    1
    >>> countBits(1000)
    10
    >>> countBits(44100)
    16
    >>> countBits(18446744073709551615)
    64
    """
    assert 0 <= value
    count = 1
    bits = 1
    while (1 << bits) <= value:
        count += bits
        value >>= bits
        bits <<= 1
    while 2 <= value:
        if bits != 1:
            bits >>= 1
        else:
            bits -= 1
        while (1 << bits) <= value:
            count += bits
            value >>= bits
    return count


def byte2bin(number, classic_mode=True):
    """
    Convert a byte (integer in 0..255 range) to a binary string.
    If classic_mode is true (default value), reverse bits.

    >>> byte2bin(10)
    '00001010'
    >>> byte2bin(10, False)
    '01010000'
    """
    text = ""
    for i in range(0, 8):
        if classic_mode:
            mask = 1 << (7 - i)
        else:
            mask = 1 << i
        if (number & mask) == mask:
            text += "1"
        else:
            text += "0"
    return text


def long2raw(value, endian, size=None):
    r"""
    Convert a number (positive and not nul) to a raw string.
    If size is given, add nul bytes to fill to size bytes.

    >>> long2raw(0x1219, BIG_ENDIAN)
    b'\x12\x19'
    >>> long2raw(0x1219, BIG_ENDIAN, 4)   # 32 bits
    b'\x00\x00\x12\x19'
    >>> long2raw(0x1219, LITTLE_ENDIAN, 4)   # 32 bits
    b'\x19\x12\x00\x00'
    """
    assert (not size and 0 < value) or (0 <= value)
    assert endian in (LITTLE_ENDIAN, BIG_ENDIAN)
    text = []
    while (value != 0 or text == ""):
        byte = value % 256
        text.append(byte)
        value >>= 8
    if size:
        need = max(size - len(text), 0)
        for i in range(need):
            text.append(0)
    if endian == BIG_ENDIAN:
        text = reversed(text)
    return bytes(text)


def long2bin(size, value, endian, classic_mode=False):
    """
    Convert a number into bits (in a string):
    - size: size in bits of the number
    - value: positive (or nul) number
    - endian: BIG_ENDIAN (most important bit first)
      or LITTLE_ENDIAN (least important bit first)
    - classic_mode (default: False): reverse each packet of 8 bits

    >>> long2bin(16, 1+4 + (1+8)*256, BIG_ENDIAN)
    '10100000 10010000'
    >>> long2bin(16, 1+4 + (1+8)*256, BIG_ENDIAN, True)
    '00000101 00001001'
    >>> long2bin(16, 1+4 + (1+8)*256, LITTLE_ENDIAN)
    '00001001 00000101'
    >>> long2bin(16, 1+4 + (1+8)*256, LITTLE_ENDIAN, True)
    '10010000 10100000'
    """
    text = ""
    assert endian in (LITTLE_ENDIAN, BIG_ENDIAN)
    assert 0 <= value
    for index in range(size):
        if (value & 1) == 1:
            text += "1"
        else:
            text += "0"
        value >>= 1
    if endian is LITTLE_ENDIAN:
        text = text[::-1]
    result = ""
    while len(text) != 0:
        if len(result) != 0:
            result += " "
        if classic_mode:
            result += text[7::-1]
        else:
            result += text[:8]
        text = text[8:]
    return result


def str2bin(value, classic_mode=True):
    r"""
    Convert binary string to binary numbers.
    If classic_mode  is true (default value), reverse bits.

    >>> str2bin("\x03\xFF")
    '00000011 11111111'
    >>> str2bin("\x03\xFF", False)
    '11000000 11111111'
    """
    text = ""
    for character in value:
        if text != "":
            text += " "
        byte = ord(character)
        text += byte2bin(byte, classic_mode)
    return text


def _createStructFormat():
    """
    Create a dictionnary (endian, size_byte) => struct format used
    by str2long() to convert raw data to positive integer.
    """
    format = {
        BIG_ENDIAN: {},
        LITTLE_ENDIAN: {},
    }
    for struct_format in "BHILQ":
        try:
            size = calcsize(struct_format)
            format[BIG_ENDIAN][size] = '>%s' % struct_format
            format[LITTLE_ENDIAN][size] = '<%s' % struct_format
        except struct_error:
            pass
    return format


_struct_format = _createStructFormat()


def str2long(data, endian):
    r"""
    Convert a raw data (type 'bytes') into a long integer.

    >>> chr(str2long(b'*', BIG_ENDIAN))
    '*'
    >>> str2long(b"\x00\x01\x02\x03", BIG_ENDIAN) == 0x10203
    True
    >>> str2long(b"\x2a\x10", LITTLE_ENDIAN) == 0x102a
    True
    >>> str2long(b"\xff\x14\x2a\x10", BIG_ENDIAN) == 0xff142a10
    True
    >>> str2long(b"\x00\x01\x02\x03", LITTLE_ENDIAN) == 0x3020100
    True
    >>> (str2long(b"\xff\x14\x2a\x10\xab\x00\xd9\x0e", BIG_ENDIAN)
    ...  == 0xff142a10ab00d90e)
    True
    >>> str2long(b"\xff\xff\xff\xff\xff\xff\xff\xff", BIG_ENDIAN) == (2**64-1)
    True
    >>> str2long(b"\x0b\x0a\x0d\x0c", MIDDLE_ENDIAN) == 0x0a0b0c0d
    True
    """
    assert 1 <= len(data) <= 32   # arbitrary limit: 256 bits
    try:
        return unpack(_struct_format[endian][len(data)], data)[0]
    except KeyError:
        pass

    assert endian in (BIG_ENDIAN, LITTLE_ENDIAN, MIDDLE_ENDIAN)
    shift = 0
    value = 0
    if endian is BIG_ENDIAN:
        data = reversed(data)
    elif endian is MIDDLE_ENDIAN:
        data = reversed(strswapmid(data))
    for byte in data:
        value += (byte << shift)
        shift += 8
    return value



"""
Moving Picture Experts Group (MPEG) video version 1 and 2 parser.

Information:
- http://www.mpucoder.com/DVD/
- http://dvd.sourceforge.net/dvdinfo/
- http://www.mit.jyu.fi/mweber/leffakone/software/parsempegts/
- http://homepage.mac.com/rnc/EditMpegHeaderIFO.html
- http://standards.iso.org/ittf/PubliclyAvailableStandards/c025029_ISO_IEC_TR_11172-5_1998(E)_Software_Simulation.zip
    This is a sample encoder/decoder implementation for MPEG-1.

Author: Victor Stinner
Creation date: 15 september 2006
"""

from hachoir.parser import Parser
from hachoir.parser.audio.mpeg_audio import MpegAudioFile
from hachoir.field import (FieldSet,
                           FieldError, ParserError,
                           Bit, Bits, Bytes, RawBits, PaddingBits, NullBits,
                           UInt8, UInt16,
                           RawBytes, PaddingBytes,
                           Enum, CustomFragment)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal


class Timestamp(FieldSet):
    static_size = 36

    def createFields(self):
        yield Bits(self, "ts_32_30", 3)
        yield Bit(self, "sync[]")  # =True
        yield Bits(self, "ts_29_15", 15)
        yield Bit(self, "sync[]")  # =True
        yield Bits(self, "ts_14_0", 15)
        yield Bit(self, "sync[]")  # =True

    def createValue(self):
        return ((self["ts_32_30"].value << 30)
                + (self["ts_29_15"].value << 15)
                + self["ts_14_0"].value)


class SCRExt(FieldSet):
    static_size = 46

    def createFields(self):
        yield Bits(self, "scr_32_30", 3)
        yield Bit(self, "sync[]")  # =True
        yield Bits(self, "scr_29_15", 15)
        yield Bit(self, "sync[]")  # =True
        yield Bits(self, "scr_14_0", 15)
        yield Bit(self, "sync[]")  # =True
        yield Bits(self, "scr_ext", 9)
        yield Bit(self, "sync[]")  # =True

    def createDescription(self):
        return "System Clock Reference (90kHz per tick), with extended 27MHz resolution"

    def createValue(self):
        return ((self["scr_32_30"].value << 30)
                + (self["scr_29_15"].value << 15)
                + self["scr_14_0"].value
                + (self["scr_ext"].value / 300.0))


class PackHeader(FieldSet):

    def createFields(self):
        if self.stream.readBits(self.absolute_address, 2, self.endian) == 1:
            # MPEG version 2
            yield Bits(self, "sync[]", 2)
            yield SCRExt(self, "scr")
            yield Bits(self, "mux_rate", 22)
            yield Bits(self, "sync[]", 2)
            yield PaddingBits(self, "reserved", 5, pattern=1)
            yield Bits(self, "stuffing_length", 3)
            count = self["stuffing_length"].value
            if count:
                yield PaddingBytes(self, "stuffing", count, pattern="\xff")
        else:
            # MPEG version 1
            yield Bits(self, "sync[]", 4)
            yield Timestamp(self, "scr")
            yield Bit(self, "sync[]")
            yield Bits(self, "mux_rate", 22)
            yield Bit(self, "sync[]")

    def validate(self):
        if self["mux_rate"].value == 0:
            return "Invalid mux rate"
        sync0 = self["sync[0]"]
        if (sync0.size == 2 and sync0.value == 1):
            # MPEG2
            if self["sync[1]"].value != 3:
                return "Invalid synchronisation bits"
        elif (sync0.size == 4 and sync0.value == 2):
            # MPEG1
            if not self["sync[1]"].value \
                    or not self["sync[2]"].value:
                return "Invalid synchronisation bits"
        else:
            return "Unknown version"
        return True


class SystemHeader(FieldSet):

    def createFields(self):
        yield Bits(self, "marker[]", 1)
        yield Bits(self, "rate_bound", 22)
        yield Bits(self, "marker[]", 1)
        yield Bits(self, "audio_bound", 6)
        yield Bit(self, "fixed_bitrate")
        yield Bit(self, "csps", description="Constrained system parameter stream")
        yield Bit(self, "audio_lock")
        yield Bit(self, "video_lock")
        yield Bits(self, "marker[]", 1)
        yield Bits(self, "video_bound", 5)
        length = self['../length'].value - 5
        if length:
            yield RawBytes(self, "raw[]", length)


class defaultParser(FieldSet):

    def createFields(self):
        yield RawBytes(self, "data", self["../length"].value)


class Padding(FieldSet):

    def createFields(self):
        yield PaddingBytes(self, "data", self["../length"].value)


class VideoExtension2(FieldSet):

    def createFields(self):
        yield Bit(self, "sync[]")  # =True
        yield Bits(self, "ext_length", 7)
        yield NullBits(self, "reserved[]", 8)
        size = self["ext_length"].value
        if size:
            yield RawBytes(self, "ext_bytes", size)


class VideoExtension1(FieldSet):

    def createFields(self):
        yield Bit(self, "has_private")
        yield Bit(self, "has_pack_lgth")
        yield Bit(self, "has_pack_seq")
        yield Bit(self, "has_pstd_buffer")
        yield Bits(self, "sync[]", 3)  # =7
        yield Bit(self, "has_extension2")

        if self["has_private"].value:
            yield RawBytes(self, "private", 16)

        if self["has_pack_lgth"].value:
            yield UInt8(self, "pack_lgth")

        if self["has_pack_seq"].value:
            yield Bit(self, "sync[]")  # =True
            yield Bits(self, "pack_seq_counter", 7)
            yield Bit(self, "sync[]")  # =True
            yield Bit(self, "mpeg12_id")
            yield Bits(self, "orig_stuffing_length", 6)

        if self["has_pstd_buffer"].value:
            yield Bits(self, "sync[]", 2)  # =1
            yield Enum(Bit(self, "pstd_buffer_scale"),
                       {True: "128 bytes", False: "1024 bytes"})
            yield Bits(self, "pstd_size", 13)


class VideoSeqHeader(FieldSet):
    ASPECT = ["forbidden", "1.0000 (VGA etc.)", "0.6735",
              "0.7031 (16:9, 625line)", "0.7615", "0.8055",
              "0.8437 (16:9, 525line)", "0.8935",
              "0.9157 (CCIR601, 625line)", "0.9815", "1.0255", "1.0695",
              "1.0950 (CCIR601, 525line)", "1.1575", "1.2015", "reserved"]
    FRAMERATE = ["forbidden", "23.976 fps", "24 fps", "25 fps", "29.97 fps",
                 "30 fps", "50 fps", "59.94 fps", "60 fps"]

    def createFields(self):
        yield Bits(self, "width", 12)
        yield Bits(self, "height", 12)
        yield Enum(Bits(self, "aspect", 4), self.ASPECT)
        yield Enum(Bits(self, "frame_rate", 4), self.FRAMERATE)
        yield Bits(self, "bit_rate", 18, "Bit rate in units of 50 bytes")
        yield Bits(self, "sync[]", 1)  # =1
        yield Bits(self, "vbv_size", 10, "Video buffer verifier size, in units of 16768")
        yield Bit(self, "constrained_params_flag")
        yield Bit(self, "has_intra_quantizer")
        if self["has_intra_quantizer"].value:
            for i in range(64):
                yield Bits(self, "intra_quantizer[]", 8)
        yield Bit(self, "has_non_intra_quantizer")
        if self["has_non_intra_quantizer"].value:
            for i in range(64):
                yield Bits(self, "non_intra_quantizer[]", 8)


class GroupStart(FieldSet):

    def createFields(self):
        yield Bit(self, "drop_frame")
        yield Bits(self, "time_hh", 5)
        yield Bits(self, "time_mm", 6)
        yield PaddingBits(self, "time_pad[]", 1)
        yield Bits(self, "time_ss", 6)
        yield Bits(self, "time_ff", 6)
        yield Bit(self, "closed_group")
        yield Bit(self, "broken_group")
        yield PaddingBits(self, "pad[]", 5)


class PacketElement(FieldSet):

    def createFields(self):
        yield Bits(self, "sync[]", 2)  # =2
        if self["sync[0]"].value != 2:
            raise ParserError("Unknown video elementary data")
        yield Bits(self, "is_scrambled", 2)
        yield Bits(self, "priority", 1)
        yield Bit(self, "alignment")
        yield Bit(self, "is_copyrighted")
        yield Bit(self, "is_original")
        yield Bit(self, "has_pts", "Presentation Time Stamp")
        yield Bit(self, "has_dts", "Decode Time Stamp")
        yield Bit(self, "has_escr", "Elementary Stream Clock Reference")
        yield Bit(self, "has_es_rate", "Elementary Stream rate")
        yield Bit(self, "dsm_trick_mode")
        yield Bit(self, "has_copy_info")
        yield Bit(self, "has_prev_crc", "If True, previous PES packet CRC follows")
        yield Bit(self, "has_extension")
        yield UInt8(self, "size")

        # Time stamps
        if self["has_pts"].value:
            yield Bits(self, "sync[]", 4)  # =2, or 3 if has_dts=True
            yield Timestamp(self, "pts")
        if self["has_dts"].value:
            if not self["has_pts"].value:
                raise ParserError("Invalid PTS/DTS values")
            yield Bits(self, "sync[]", 4)  # =1
            yield Timestamp(self, "dts")

        if self["has_escr"].value:
            yield Bits(self, "sync[]", 2)  # =0
            yield SCRExt(self, "escr")

        if self["has_es_rate"].value:
            yield Bit(self, "sync[]")  # =True
            yield Bits(self, "es_rate", 14)  # in units of 50 bytes/second
            yield Bit(self, "sync[]")  # =True

        if self["has_copy_info"].value:
            yield Bit(self, "sync[]")  # =True
            yield Bits(self, "copy_info", 7)

        if self["has_prev_crc"].value:
            yield textHandler(UInt16(self, "prev_crc"), hexadecimal)

        # --- Extension ---
        if self["has_extension"].value:
            yield VideoExtension1(self, "extension")
            if self["extension/has_extension2"].value:
                yield VideoExtension2(self, "extension2")


class VideoExtension(FieldSet):
    EXT_TYPE = {1: 'Sequence', 2: 'Sequence Display', 8: 'Picture Coding'}

    def createFields(self):
        yield Enum(Bits(self, "ext_type", 4), self.EXT_TYPE)
        ext_type = self['ext_type'].value
        if ext_type == 1:
            # Sequence extension
            yield Bits(self, 'profile_and_level', 8)
            yield Bit(self, 'progressive_sequence')
            yield Bits(self, 'chroma_format', 2)
            yield Bits(self, 'horiz_size_ext', 2)
            yield Bits(self, 'vert_size_ext', 2)
            yield Bits(self, 'bit_rate_ext', 12)
            yield Bits(self, 'pad[]', 1)
            yield Bits(self, 'vbv_buffer_size_ext', 8)
            yield Bit(self, 'low_delay')
            yield Bits(self, 'frame_rate_ext_n', 2)
            yield Bits(self, 'frame_rate_ext_d', 5)
        elif ext_type == 2:
            # Sequence Display extension
            yield Bits(self, 'video_format', 3)
            yield Bit(self, 'color_desc_present')
            if self['color_desc_present'].value:
                yield UInt8(self, 'color_primaries')
                yield UInt8(self, 'transfer_characteristics')
                yield UInt8(self, 'matrix_coeffs')
            yield Bits(self, 'display_horiz_size', 14)
            yield Bits(self, 'pad[]', 1)
            yield Bits(self, 'display_vert_size', 14)
            yield NullBits(self, 'pad[]', 3)
        elif ext_type == 8:
            yield Bits(self, 'f_code[0][0]', 4, description="forward horizontal")
            yield Bits(self, 'f_code[0][1]', 4, description="forward vertical")
            yield Bits(self, 'f_code[1][0]', 4, description="backward horizontal")
            yield Bits(self, 'f_code[1][1]', 4, description="backward vertical")
            yield Bits(self, 'intra_dc_precision', 2)
            yield Bits(self, 'picture_structure', 2)
            yield Bit(self, 'top_field_first')
            yield Bit(self, 'frame_pred_frame_dct')
            yield Bit(self, 'concealment_motion_vectors')
            yield Bit(self, 'q_scale_type')
            yield Bit(self, 'intra_vlc_format')
            yield Bit(self, 'alternate_scan')
            yield Bit(self, 'repeat_first_field')
            yield Bit(self, 'chroma_420_type')
            yield Bit(self, 'progressive_frame')
            yield Bit(self, 'composite_display')
            if self['composite_display'].value:
                yield Bit(self, 'v_axis')
                yield Bits(self, 'field_sequence', 3)
                yield Bit(self, 'sub_carrier')
                yield Bits(self, 'burst_amplitude', 7)
                yield Bits(self, 'sub_carrier_phase', 8)
                yield NullBits(self, 'pad[]', 2)
            else:
                yield NullBits(self, 'pad[]', 6)
        else:
            yield RawBits(self, "raw[]", 4)


class VideoPicture(FieldSet):
    CODING_TYPE = ["forbidden", "intra-coded (I)",
                   "predictive-coded (P)",
                   "bidirectionally-predictive-coded (B)",
                   "dc intra-coded (D)", "reserved",
                   "reserved", "reserved"]

    def createFields(self):
        yield Bits(self, "temporal_ref", 10)
        yield Enum(Bits(self, "coding_type", 3), self.CODING_TYPE)
        yield Bits(self, "vbv_delay", 16)
        if self['coding_type'].value in (2, 3):
            # predictive coding
            yield Bit(self, 'full_pel_fwd_vector')
            yield Bits(self, 'forward_f_code', 3)
        if self['coding_type'].value == 3:
            # bidi predictive coding
            yield Bit(self, 'full_pel_back_vector')
            yield Bits(self, 'backward_f_code', 3)
        yield Bits(self, "padding", 8 - (self.current_size % 8))


class VideoSlice(FieldSet):

    def createFields(self):
        yield Bits(self, "quantizer_scale", 5)
        start = self.absolute_address + self.current_size + 3
        # seek forward by at most 1MB
        pos = self.stream.searchBytes(
            b'\0\0\1', start, start + 1024 * 1024 * 8)
        if pos is None:
            pos = self.root.size
        yield RawBits(self, "data", pos - start + 3)


class VideoChunk(FieldSet):
    tag_info = {
        0x00: ("pict_start[]", VideoPicture, "Picture start"),
        0xB2: ("data_start[]", None, "Data start"),
        0xB3: ("seq_hdr[]", VideoSeqHeader, "Sequence header"),
        0xB4: ("seq_err[]", None, "Sequence error"),
        0xB5: ("ext_start[]", VideoExtension, "Extension start"),
        0xB7: ("seq_end[]", None, "Sequence end"),
        0xB8: ("group_start[]", GroupStart, "Group start"),
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        tag = self["tag"].value
        if tag in self.tag_info:
            self._name, self.parser, self._description = self.tag_info[tag]
            if not self.parser:
                self.parser = defaultParser
        elif 0x01 <= tag <= 0xaf:
            self._name, self.parser, self._description = (
                'slice[]', VideoSlice, 'Picture slice')
        else:
            self.parser = defaultParser

    def createFields(self):
        yield Bytes(self, "sync", 3)
        yield textHandler(UInt8(self, "tag"), hexadecimal)
        if self.parser and self['tag'].value != 0xb7:
            yield self.parser(self, "content")


class VideoStream(Parser):
    endian = BIG_ENDIAN

    def createFields(self):
        while self.current_size < self.size:
            # seek forward by at most 1MB
            pos = self.stream.searchBytes(
                b'\0\0\1', self.current_size, self.current_size + 1024 * 1024 * 8)
            if pos is not None:
                padsize = pos - self.current_size
                if padsize:
                    yield PaddingBytes(self, "pad[]", padsize // 8)
            yield VideoChunk(self, "chunk[]")


class Stream(FieldSet):

    def createFields(self):
        padding = 0
        position = 0
        streamlength = self["../length"].value
        while position < streamlength * 8:
            next = ord(self.parent.stream.readBytes(
                self.absolute_address + self.current_size + position, 1))
            if next == 0xff:
                padding += 1
                position += 8
            elif padding:
                yield PaddingBytes(self, "pad[]", padding)
                padding = None
                position = 0
            elif 0x40 <= next <= 0x7f:
                yield Bits(self, "scale_marker", 2)  # 1
                yield Bit(self, "scale")
                scale = self['scale'].value
                if scale:
                    scaleval = 1024
                else:
                    scaleval = 128
                yield textHandler(Bits(self, "size", 13), lambda field: str(field.value * scaleval))
            elif 0x00 <= next <= 0x3f:
                yield Bits(self, "ts_marker", 2)  # 0
                yield Bit(self, "has_pts")
                yield Bit(self, "has_dts")
                if self['has_pts'].value:
                    yield Timestamp(self, "pts")
                if self['has_dts'].value:
                    yield PaddingBits(self, "pad[]", 4)
                    yield Timestamp(self, "dts")
                if self.current_size % 8 == 4:
                    yield PaddingBits(self, "pad[]", 4)
                break
            elif 0x80 <= next <= 0xbf:
                # MPEG-2 extension
                yield PacketElement(self, "pkt")
                break
            else:
                # 0xc0 - 0xfe: unknown
                break
        length = self["../length"].value - self.current_size // 8
        if length:
            tag = self['../tag'].value
            group = self.root.streamgroups[tag]
            parname = self.parent._name
            if parname.startswith('audio'):
                frag = CustomFragment(
                    self, "data", length * 8, MpegAudioFile, group=group)
            elif parname.startswith('video'):
                frag = CustomFragment(
                    self, "data", length * 8, VideoStream, group=group)
            else:
                frag = CustomFragment(
                    self, "data", length * 8, None, group=group)
            self.root.streamgroups[tag] = frag.group
            yield frag


class Chunk(FieldSet):
    ISO_END_CODE = 0xB9
    tag_info = {
        0xB9: ("end", None, "End"),
        0xBA: ("pack_start[]", PackHeader, "Pack start"),
        0xBB: ("system_start[]", SystemHeader, "System start"),
        # streams
        0xBD: ("private[]", Stream, "Private elementary"),
        0xBE: ("padding[]", Stream, "Padding"),
        # 0xC0 to 0xFE handled specially
        0xFF: ("directory[]", Stream, "Program Stream Directory"),
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        if not hasattr(self.root, 'streamgroups'):
            self.root.streamgroups = {}
            for tag in range(0xBC, 0x100):
                self.root.streamgroups[tag] = None
        tag = self["tag"].value
        if tag in self.tag_info:
            self._name, self.parser, self._description = self.tag_info[tag]
        elif 0xBC <= tag <= 0xFF:
            if 0xC0 <= tag < 0xE0:
                # audio
                streamid = tag - 0xC0
                self._name, self.parser, self._description = (
                    "audio[%i][]" % streamid, Stream, "Audio Stream %i Packet" % streamid)
            elif 0xE0 <= tag < 0xF0:
                # video
                streamid = tag - 0xE0
                self._name, self.parser, self._description = (
                    "video[%i][]" % streamid, Stream, "Video Stream %i Packet" % streamid)
            else:
                self._name, self.parser, self._description = (
                    "stream[]", Stream, "Data Stream Packet")
        else:
            self.parser = defaultParser

        if not self.parser:
            self.parser = defaultParser
        elif self.parser != PackHeader and "length" in self:
            self._size = (6 + self["length"].value) * 8

    def createFields(self):
        yield Bytes(self, "sync", 3)
        yield textHandler(UInt8(self, "tag"), hexadecimal)
        if self.parser:
            if self.parser != PackHeader:
                yield UInt16(self, "length")
                if not self["length"].value:
                    return
            yield self.parser(self, "content")

    def createDescription(self):
        return "Chunk: tag %s" % self["tag"].display


class MPEGVideoFile(Parser):
    PARSER_TAGS = {
        "id": "mpeg_video",
        "category": "video",
        "file_ext": ("mpeg", "mpg", "mpe", "vob"),
        "mime": ("video/mpeg", "video/mp2p"),
        "min_size": 12 * 8,
        # TODO:        "magic": xxx,
        "description": "MPEG video, version 1 or 2"
    }
    endian = BIG_ENDIAN
    version = None

    def createFields(self):
        while self.current_size < self.size:
            # seek forward by at most 1MB
            pos = self.stream.searchBytes(
                b'\0\0\1', self.current_size, self.current_size + 1024 * 1024 * 8)
            if pos is not None:
                padsize = pos - self.current_size
                if padsize:
                    yield PaddingBytes(self, "pad[]", padsize // 8)
            chunk = Chunk(self, "chunk[]")
            try:
                # force chunk to be processed, so that CustomFragments are
                # complete
                chunk['content/data']
            except Exception:
                pass
            yield chunk

    def validate(self):
        try:
            pack = self[0]
        except FieldError:
            return "Unable to create first chunk"
        if pack.name != "pack_start[0]":
            return "Invalid first chunk"
        if pack["sync"].value != b"\0\0\1":
            return "Invalid synchronisation"
        return pack["content"].validate()

    def getVersion(self):
        if not self.version:
            if self["pack_start[0]/content/sync[0]"].size == 2:
                self.version = 2
            else:
                self.version = 1
        return self.version

    def createDescription(self):
        if self.getVersion() == 2:
            return "MPEG-2 video"
        else:
            return "MPEG-1 video"



def Enum(field, enum, key_func=None):
    """
    Enum is an adapter to another field: it will just change its display
    attribute. It uses a dictionary to associate a value to another.

    key_func is an optional function with prototype "def func(key)->key"
    which is called to transform key.
    """
    display = field.createDisplay
    if key_func:
        def createDisplay():
            try:
                key = key_func(field.value)
                return enum[key]
            except LookupError:
                return display()
    else:
        def createDisplay():
            try:
                return enum[field.value]
            except LookupError:
                return display()
    field.createDisplay = createDisplay
    field.getEnum = lambda: enum
    return field



"""
RealAudio (.ra) parser

Author: Mike Melanson
References:
  http://wiki.multimedia.cx/index.php?title=RealMedia
Samples:
  http://samples.mplayerhq.hu/real/RA/
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt8, UInt16, UInt32,
                           Bytes, RawBytes, String,
                           PascalString8)
from hachoir.core.tools import humanFrequency
from hachoir.core.text_handler import displayHandler
from hachoir.core.endian import BIG_ENDIAN


class Metadata(FieldSet):

    def createFields(self):
        yield PascalString8(self, "title", charset="ISO-8859-1")
        yield PascalString8(self, "author", charset="ISO-8859-1")
        yield PascalString8(self, "copyright", charset="ISO-8859-1")
        yield PascalString8(self, "comment", charset="ISO-8859-1")


class RealAudioFile(Parser):
    MAGIC = b".ra\xFD"
    PARSER_TAGS = {
        "id": "real_audio",
        "category": "audio",
        "file_ext": ["ra"],
        "mime": ("audio/x-realaudio", "audio/x-pn-realaudio"),
        "min_size": 6 * 8,
        "magic": ((MAGIC, 0),),
        "description": "Real audio (.ra)",
    }
    endian = BIG_ENDIAN

    def validate(self):
        if self["signature"].value != self.MAGIC:
            return "Invalid signature"
        if self["version"].value not in (3, 4):
            return "Unknown version"
        return True

    def createFields(self):
        yield Bytes(self, "signature", 4, r"RealAudio identifier ('.ra\xFD')")
        yield UInt16(self, "version", "Version")
        if self["version"].value == 3:
            yield UInt16(self, "header_size", "Header size")
            yield RawBytes(self, "Unknown1", 10)
            yield UInt32(self, "data_size", "Data size")
            yield Metadata(self, "metadata")
            yield UInt8(self, "Unknown2")
            yield PascalString8(self, "FourCC")
            audio_size = self["data_size"].value
        else:  # version = 4
            yield UInt16(self, "reserved1", "Reserved, should be 0")
            yield String(self, "ra4sig", 4, "'.ra4' signature")
            yield UInt32(self, "filesize", "File size (minus 40 bytes)")
            yield UInt16(self, "version2", "Version 2 (always equal to version)")
            yield UInt32(self, "headersize", "Header size (minus 16)")
            yield UInt16(self, "codec_flavor", "Codec flavor")
            yield UInt32(self, "coded_frame_size", "Coded frame size")
            yield RawBytes(self, "unknown1", 12)
            yield UInt16(self, "subpacketh", "Subpacket h (?)")
            yield UInt16(self, "frame_size", "Frame size")
            yield UInt16(self, "sub_packet_size", "Subpacket size")
            yield UInt16(self, "unknown2", "Unknown")
            yield displayHandler(UInt16(self, "sample_rate", "Sample rate"), humanFrequency)
            yield UInt16(self, "unknown3", "Unknown")
            yield UInt16(self, "sample_size", "Sample size")
            yield UInt16(self, "channels", "Channels")
            yield PascalString8(self, "Interleaving ID String")
            yield PascalString8(self, "FourCC")
            yield RawBytes(self, "unknown4", 3)
            yield Metadata(self, "metadata")
            audio_size = (self["filesize"].value + 40) - \
                (self["headersize"].value + 16)
        if 0 < audio_size:
            yield RawBytes(self, "audio_data", audio_size)

    def createDescription(self):
        if (self["version"].value == 3):
            return "RealAudio v3 file, '%s' codec" % self["FourCC"].value
        elif (self["version"].value == 4):
            return "RealAudio v4 file, '%s' codec, %s, %u channels" % (
                self["FourCC"].value, self["sample_rate"].display, self["channels"].value)
        else:
            return "Real audio"



"""
Audio Interchange File Format (AIFF) parser.

Author: Victor Stinner
Creation: 27 december 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt16, UInt32, Float80, TimestampMac32,
                           RawBytes, NullBytes,
                           String, Enum, PascalString32)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import filesizeHandler
from hachoir.core.tools import alignValue
from hachoir.parser.audio.id3 import ID3v2

CODEC_NAME = {
    'ACE2': "ACE 2-to-1",
    'ACE8': "ACE 8-to-3",
    'MAC3': "MAC 3-to-1",
    'MAC6': "MAC 6-to-1",
    'NONE': "None",
    'sowt': "Little-endian, no compression",
}


class Comment(FieldSet):

    def createFields(self):
        yield TimestampMac32(self, "timestamp")
        yield PascalString32(self, "text")


def parseText(self):
    yield String(self, "text", self["size"].value)


def parseID3(self):
    yield ID3v2(self, "id3v2", size=self["size"].value * 8)


def parseComment(self):
    yield UInt16(self, "nb_comment")
    for index in range(self["nb_comment"].value):
        yield Comment(self, "comment[]")


def parseCommon(self):
    yield UInt16(self, "nb_channel")
    yield UInt32(self, "nb_sample")
    yield UInt16(self, "sample_size")
    yield Float80(self, "sample_rate")
    yield Enum(String(self, "codec", 4, strip="\0", charset="ASCII"), CODEC_NAME)


def parseVersion(self):
    yield TimestampMac32(self, "timestamp")


def parseSound(self):
    yield UInt32(self, "offset")
    yield UInt32(self, "block_size")
    size = (self.size - self.current_size) // 8
    if size:
        yield RawBytes(self, "data", size)


class Chunk(FieldSet):
    TAG_INFO = {
        'COMM': ('common', "Common chunk", parseCommon),
        'COMT': ('comment', "Comment", parseComment),
        'NAME': ('name', "Name", parseText),
        'AUTH': ('author', "Author", parseText),
        'FVER': ('version', "Version", parseVersion),
        'SSND': ('sound', "Sound data", parseSound),
        'ID3 ': ('id3', "ID3", parseID3),
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = (8 + alignValue(self["size"].value, 2)) * 8
        tag = self["type"].value
        if tag in self.TAG_INFO:
            self._name, self._description, self._parser = self.TAG_INFO[tag]
        else:
            self._parser = None

    def createFields(self):
        yield String(self, "type", 4, "Signature (FORM)", charset="ASCII")
        yield filesizeHandler(UInt32(self, "size"))
        size = self["size"].value
        if size:
            if self._parser:
                yield from self._parser(self)
                if size % 2:
                    yield NullBytes(self, "padding", 1)
            else:
                yield RawBytes(self, "data", size)


class AiffFile(Parser):
    PARSER_TAGS = {
        "id": "aiff",
        "category": "audio",
        "file_ext": ("aif", "aiff", "aifc"),
        "mime": ("audio/x-aiff",),
        "magic_regex": ((b"FORM.{4}AIF[CF]", 0),),
        "min_size": 12 * 8,
        "description": "Audio Interchange File Format (AIFF)"
    }
    endian = BIG_ENDIAN

    def validate(self):
        if self.stream.readBytes(0, 4) != b"FORM":
            return "Invalid signature"
        if self.stream.readBytes(8 * 8, 4) not in (b"AIFF", b"AIFC"):
            return "Invalid type"
        return True

    def createFields(self):
        yield String(self, "signature", 4, "Signature (FORM)", charset="ASCII")
        yield filesizeHandler(UInt32(self, "filesize"))
        yield String(self, "type", 4, "Form type (AIFF or AIFC)", charset="ASCII")
        while not self.eof:
            yield Chunk(self, "chunk[]")

    def createDescription(self):
        if self["type"].value == "AIFC":
            return "Audio Interchange File Format Compressed (AIFC)"
        else:
            return "Audio Interchange File Format (AIFF)"

    def createContentSize(self):
        return self["filesize"].value * 8



from wx import ListCtrl, EVT_WINDOW_CREATE, CallAfter
from wx.lib.mixins.listctrl import ListCtrlAutoWidthMixin


class field_view_t(ListCtrl, ListCtrlAutoWidthMixin):
    def __init__(self):
        self.cols = {}

        ListCtrl.__init__(self)
        self.Bind(EVT_WINDOW_CREATE, self.on_create)

    def post_init(self):
        ListCtrlAutoWidthMixin.__init__(self)

        columns = ['address', 'name', 'type', 'size', 'data', 'description']
        for name in columns:
            self.append_column(name)
        self.col_min_width = [len(s) for s in columns]

        self.Layout()
        self.dispatcher.trigger('field_view_ready', self)

    def on_create(self, event):
        self.Unbind(EVT_WINDOW_CREATE)
        CallAfter(self.post_init)

    def append_column(self, name):
        index = self.GetColumnCount()
        self.cols[name] = index
        self.InsertColumn(col=index, heading=name)

    def get_selected(self, name):
        return self.GetItem(self.GetFocusedItem(), self.cols['name']).GetText()

    def clear(self):
        self.DeleteAllItems()

    def register_callback(self, cbGetItemText):
        self.OnGetItemText_imp = cbGetItemText

    def OnGetItemText(self, item, col):
        return self.OnGetItemText_imp(item, col)

    def get_col_index(self, name):
        return self.cols[name]

    def get_col_count(self):
        return len(self.cols)

    def resize_column(self, col_index, width):
        width = max(self.col_min_width[col_index], width) + 1
        self.SetColumnWidth(col_index, self.GetCharWidth() * width)



# from hachoir.wx.field_view.mutator import split_field
from hachoir.field import RawBytes, RawBits


class field_split_menu_imp_t:
    def on_field_split_menu_ready(self, dispatcher, view):
        assert view is not None
        self.view = view

    def on_field_selected(self, dispatcher, field):
        self.field = field

    def on_split_bytes(self):
        if self.split_field('Split Bytes...', self.field, RawBytes, lambda field: field.size // 8):
            self.dispatcher.trigger('field_was_split_bytes', self.field)

    def on_split_bits(self):
        if self.split_field('Split Bits...', self.field, RawBits, lambda field: field.size):
            self.dispatcher.trigger('field_was_split_bits', self.field)

    def split_field(self, caption, field, split_type, size_func):
        offset = self.view.ask_split(caption, 1, size_func(field) - 1)
        # FIXME: code commented because of pep8 warning
        # if offset is not None:
        #     new_fields = split_field(field, offset, field.name, split_type, size_func)
        return offset



"""
Garmin fit file Format parser.

Author: Sebastien Ponce <sebastien.ponce@cern.ch>
"""

from hachoir.parser import Parser
from hachoir.field import FieldSet, Int8, UInt8, Int16, UInt16, Int32, UInt32, Int64, UInt64, RawBytes, Bit, Bits, Bytes, String, Float32, Float64
from hachoir.core.endian import BIG_ENDIAN, LITTLE_ENDIAN

field_types = {
    0: UInt8,      # enum
    1: Int8,       # signed int of 8 bits
    2: UInt8,      # unsigned int of 8 bits
    131: Int16,    # signed int of 16 bits
    132: UInt16,   # unsigned int of 16 bits
    133: Int32,    # signed int of 32 bits
    134: UInt32,   # unsigned int of 32 bits
    7: String,     # string
    136: Float32,  # float
    137: Float64,  # double
    10: UInt8,     # unsigned int of 8 bits with 0 as invalid value
    139: UInt16,   # unsigned int of 16 bits with 0 as invalid value
    140: UInt32,   # unsigned int of 32 bits with 0 as invalid value
    13: Bytes,     # bytes
    142: Int64,    # signed int of 64 bits
    143: UInt64,   # unsigned int of 64 bits
    144: UInt64    # unsigned int of 64 bits with 0 as invalid value
}


class Header(FieldSet):
    endian = LITTLE_ENDIAN

    def createFields(self):
        yield UInt8(self, "size", "Header size")
        yield UInt8(self, "protocol", "Protocol version")
        yield UInt16(self, "profile", "Profile version")
        yield UInt32(self, "datasize", "Data size")
        yield RawBytes(self, "datatype", 4)
        yield UInt16(self, "crc", "CRC of first 11 bytes or 0x0")

    def createDescription(self):
        return "Header of fit file. Data size is %d" % (self["datasize"].value)


class NormalRecordHeader(FieldSet):

    def createFields(self):
        yield Bit(self, "normal", "Normal header (0)")
        yield Bit(self, "type", "Message type (0 data, 1 definition")
        yield Bit(self, "typespecific", "0")
        yield Bit(self, "reserved", "0")
        yield Bits(self, "msgType", 4, description="Message type")

    def createDescription(self):
        return "Record header, this is a %s message" % ("definition" if self["type"].value else "data")


class FieldDefinition(FieldSet):

    def createFields(self):
        yield UInt8(self, "number", "Field definition number")
        yield UInt8(self, "size", "Size in bytes")
        yield UInt8(self, "type", "Base type")

    def createDescription(self):
        return "Field Definition. Number %d, Size %d" % (self["number"].value, self["size"].value)


class DefinitionMessage(FieldSet):

    def createFields(self):
        yield NormalRecordHeader(self, "RecordHeader")
        yield UInt8(self, "reserved", "Reserved (0)")
        yield UInt8(self, "architecture", "Architecture (0 little, 1 big endian")
        self.endian = BIG_ENDIAN if self["architecture"].value else LITTLE_ENDIAN
        yield UInt16(self, "msgNumber", "Message Number")
        yield UInt8(self, "nbFields", "Number of fields")
        for n in range(self["nbFields"].value):
            yield FieldDefinition(self, "fieldDefinition[]")

    def createDescription(self):
        return "Definition Message. Contains %d fields" % (self["nbFields"].value)


class DataMessage(FieldSet):

    def createFields(self):
        hdr = NormalRecordHeader(self, "RecordHeader")
        yield hdr
        msgType = self["RecordHeader"]["msgType"].value
        msgDef = self.parent.msgDefs[msgType]
        for n in range(msgDef["nbFields"].value):
            desc = msgDef["fieldDefinition[%d]" % n]
            typ = field_types[desc["type"].value]
            self.endian = BIG_ENDIAN if msgDef["architecture"].value else LITTLE_ENDIAN
            if typ == String or typ == Bytes:
                yield typ(self, "field%d" % n, desc["size"].value)
            else:
                if typ.static_size // 8 == desc["size"].value:
                    yield typ(self, "field%d" % n, desc["size"].value)
                else:
                    for p in range(desc["size"].value * 8 // typ.static_size):
                        yield typ(self, "field%d[]" % n)

    def createDescription(self):
        return "Data Message"


class TimeStamp(FieldSet):

    def createFields(self):
        yield Bit(self, "timestamp", "TimeStamp (1)")
        yield Bits(self, "msgType", 3, description="Message type")
        yield Bits(self, "time", 4, description="TimeOffset")

    def createDescription(self):
        return "TimeStamp"


class CRC(FieldSet):

    def createFields(self):
        yield UInt16(self, "crc", "CRC")

    def createDescription(self):
        return "CRC"


class FITFile(Parser):
    endian = BIG_ENDIAN
    PARSER_TAGS = {
        "id": "fit",
        "category": "misc",
        "file_ext": ("fit",),
        "mime": ("application/fit",),
        "min_size": 14 * 8,
        "description": "Garmin binary fit format"
    }

    def __init__(self, *args, **kwargs):
        Parser.__init__(self, *args, **kwargs)
        self.msgDefs = {}

    def validate(self):
        s = self.stream.readBytes(0, 12)
        if s[8:12] != b'.FIT':
            return "Invalid header %d %d %d %d" % tuple([int(b) for b in s[8:12]])
        return True

    def createFields(self):
        yield Header(self, "header")
        while self.current_size < self["header"]["datasize"].value * 8:
            b = self.stream.readBits(self.absolute_address + self.current_size, 2, self.endian)
            if b == 1:
                defMsg = DefinitionMessage(self, "definition[]")
                msgType = defMsg["RecordHeader"]["msgType"].value
                sizes = ''
                ts = 0
                for n in range(defMsg["nbFields"].value):
                    fname = "fieldDefinition[%d]" % n
                    size = defMsg[fname]["size"].value
                    ts += size
                    sizes += "%d/" % size
                sizes += "%d" % ts
                self.msgDefs[msgType] = defMsg
                yield defMsg
            elif b == 0:
                yield DataMessage(self, "data[]")
            else:
                yield TimeStamp(self, "timestamp[]")
        yield CRC(self, "crc")



"""
Parser for resource of Microsoft Windows Portable Executable (PE).

Documentation:
- Wine project
  VS_FIXEDFILEINFO structure, file include/winver.h

Author: Victor Stinner
Creation date: 2007-01-19
"""

from hachoir.field import (FieldSet, ParserError, Enum,
                           Bit, Bits, SeekableFieldSet,
                           UInt16, UInt32, TimestampUnix32,
                           Bytes, RawBytes, PaddingBytes, NullBytes, NullBits,
                           CString, String)
from hachoir.core.text_handler import textHandler, filesizeHandler, hexadecimal
from hachoir.core.tools import createDict, paddingSize, alignValue, makePrintable
from hachoir.parser.common.win32 import BitmapInfoHeader

MAX_DEPTH = 5
MAX_INDEX_PER_HEADER = 300
MAX_NAME_PER_HEADER = MAX_INDEX_PER_HEADER


class Version(FieldSet):
    static_size = 32

    def createFields(self):
        yield textHandler(UInt16(self, "minor", "Minor version number"), hexadecimal)
        yield textHandler(UInt16(self, "major", "Major version number"), hexadecimal)

    def createValue(self):
        return self["major"].value + float(self["minor"].value) / 10000


MAJOR_OS_NAME = {
    1: "DOS",
    2: "OS/2 16-bit",
    3: "OS/2 32-bit",
    4: "Windows NT",
}

MINOR_OS_BASE = 0
MINOR_OS_NAME = {
    0: "Base",
    1: "Windows 16-bit",
    2: "Presentation Manager 16-bit",
    3: "Presentation Manager 32-bit",
    4: "Windows 32-bit",
}

FILETYPE_DRIVER = 3
FILETYPE_FONT = 4
FILETYPE_NAME = {
    1: "Application",
    2: "DLL",
    3: "Driver",
    4: "Font",
    5: "VXD",
    7: "Static library",
}

DRIVER_SUBTYPE_NAME = {
    1: "Printer",
    2: "Keyboard",
    3: "Language",
    4: "Display",
    5: "Mouse",
    6: "Network",
    7: "System",
    8: "Installable",
    9: "Sound",
    10: "Communications",
}

FONT_SUBTYPE_NAME = {
    1: "Raster",
    2: "Vector",
    3: "TrueType",
}


class VersionInfoBinary(FieldSet):

    def createFields(self):
        yield textHandler(UInt32(self, "magic", "File information magic (0xFEEF04BD)"), hexadecimal)
        if self["magic"].value != 0xFEEF04BD:
            raise ParserError("EXE resource: invalid file info magic")
        yield Version(self, "struct_ver", "Structure version (1.0)")
        yield Version(self, "file_ver_ms", "File version MS")
        yield Version(self, "file_ver_ls", "File version LS")
        yield Version(self, "product_ver_ms", "Product version MS")
        yield Version(self, "product_ver_ls", "Product version LS")
        yield textHandler(UInt32(self, "file_flags_mask"), hexadecimal)

        yield Bit(self, "debug")
        yield Bit(self, "prerelease")
        yield Bit(self, "patched")
        yield Bit(self, "private_build")
        yield Bit(self, "info_inferred")
        yield Bit(self, "special_build")
        yield NullBits(self, "reserved", 26)

        yield Enum(textHandler(UInt16(self, "file_os_major"), hexadecimal), MAJOR_OS_NAME)
        yield Enum(textHandler(UInt16(self, "file_os_minor"), hexadecimal), MINOR_OS_NAME)
        yield Enum(textHandler(UInt32(self, "file_type"), hexadecimal), FILETYPE_NAME)
        field = textHandler(UInt32(self, "file_subfile"), hexadecimal)
        if field.value == FILETYPE_DRIVER:
            field = Enum(field, DRIVER_SUBTYPE_NAME)
        elif field.value == FILETYPE_FONT:
            field = Enum(field, FONT_SUBTYPE_NAME)
        yield field
        yield TimestampUnix32(self, "date_ms")
        yield TimestampUnix32(self, "date_ls")


class VersionInfoNode(FieldSet):
    TYPE_STRING = 1
    TYPE_NAME = {
        0: "binary",
        1: "string",
    }

    def __init__(self, parent, name, is_32bit=True):
        FieldSet.__init__(self, parent, name)
        self._size = alignValue(self["size"].value, 4) * 8
        self.is_32bit = is_32bit

    def createFields(self):
        yield UInt16(self, "size", "Node size (in bytes)")
        yield UInt16(self, "data_size")
        yield Enum(UInt16(self, "type"), self.TYPE_NAME)
        yield CString(self, "name", charset="UTF-16-LE")

        size = paddingSize(self.current_size // 8, 4)
        if size:
            yield NullBytes(self, "padding[]", size)
        size = self["data_size"].value
        if size:
            if self["type"].value == self.TYPE_STRING:
                if self.is_32bit:
                    size *= 2
                yield String(self, "value", size, charset="UTF-16-LE", truncate="\0")
            elif self["name"].value == "VS_VERSION_INFO":
                yield VersionInfoBinary(self, "value", size=size * 8)
                if self["value/file_flags_mask"].value == 0:
                    self.is_32bit = False
            else:
                yield RawBytes(self, "value", size)
        while 12 <= (self.size - self.current_size) // 8:
            yield VersionInfoNode(self, "node[]", self.is_32bit)
        size = (self.size - self.current_size) // 8
        if size:
            yield NullBytes(self, "padding[]", size)

    def createDescription(self):
        text = "Version info node: %s" % self["name"].value
        if self["type"].value == self.TYPE_STRING and "value" in self:
            text += "=%s" % self["value"].value
        return text


def parseVersionInfo(parent):
    yield VersionInfoNode(parent, "node[]")


def parseIcon(parent):
    yield BitmapInfoHeader(parent, "bmp_header")
    size = (parent.size - parent.current_size) // 8
    if size:
        yield RawBytes(parent, "raw", size)


class WindowsString(FieldSet):

    def createFields(self):
        yield UInt16(self, "length", "Number of 16-bit characters")
        size = self["length"].value * 2
        if size:
            yield String(self, "text", size, charset="UTF-16-LE")

    def createValue(self):
        if "text" in self:
            return self["text"].value
        else:
            return ""

    def createDisplay(self):
        return makePrintable(self.value, "UTF-8", quote='"')


def parseStringTable(parent):
    while not parent.eof:
        yield WindowsString(parent, "string[]")


RESOURCE_TYPE = {
    1: ("cursor[]", "Cursor", None),
    2: ("bitmap[]", "Bitmap", None),
    3: ("icon[]", "Icon", parseIcon),
    4: ("menu[]", "Menu", None),
    5: ("dialog[]", "Dialog", None),
    6: ("string_table[]", "String table", parseStringTable),
    7: ("font_dir[]", "Font directory", None),
    8: ("font[]", "Font", None),
    9: ("accelerators[]", "Accelerators", None),
    10: ("raw_res[]", "Unformatted resource data", None),
    11: ("message_table[]", "Message table", None),
    12: ("group_cursor[]", "Group cursor", None),
    14: ("group_icon[]", "Group icon", None),
    16: ("version_info", "Version information", parseVersionInfo),
}


class Entry(FieldSet):
    static_size = 16 * 8

    def __init__(self, parent, name, inode=None):
        FieldSet.__init__(self, parent, name)
        self.inode = inode

    def createFields(self):
        yield textHandler(UInt32(self, "rva"), hexadecimal)
        yield filesizeHandler(UInt32(self, "size"))
        yield UInt32(self, "codepage")
        yield NullBytes(self, "reserved", 4)

    def createDescription(self):
        return "Entry #%u: offset=%s size=%s" % (
            self.inode["offset"].value, self["rva"].display, self["size"].display)


class NameOffset(FieldSet):

    def __init__(self, parent, name):
        FieldSet.__init__(self, parent, name)
        self.name_field = None

    def createFields(self):
        yield Bits(self, "name_offset", 31)
        yield Bit(self, "is_name")
        yield Bits(self, "offset", 31)
        yield Bit(self, "is_subdir")

    def getResType(self):
        return self.name_field.value

    def createDescription(self):
        if self["is_subdir"].value:
            return "Sub-directory: %s at %s" % (self.name_field.display, self["offset"].value)
        else:
            return "Index: %s at %s" % (self.name_field.display, self["offset"].value)


class IndexOffset(FieldSet):
    TYPE_DESC = createDict(RESOURCE_TYPE, 1)

    def __init__(self, parent, name, res_type=None):
        FieldSet.__init__(self, parent, name)
        self.res_type = res_type

    def createFields(self):
        if self.res_type is None:
            # immediate subdirectory of the root
            yield Enum(UInt32(self, "type"), self.TYPE_DESC)
        else:
            # sub-subdirectory, "type" field is just an ID
            yield textHandler(UInt32(self, "type"), lambda field: "ID %d" % field.value)
        yield Bits(self, "offset", 31)
        yield Bit(self, "is_subdir")

    def getResType(self):
        return self["type"].value

    def createDescription(self):
        if self["is_subdir"].value:
            return "Sub-directory: %s at %s" % (self["type"].display, self["offset"].value)
        else:
            return "Index: %s at %s" % (self["type"].display, self["offset"].value)


class ResourceContent(FieldSet):

    def __init__(self, parent, name, entry, size=None):
        FieldSet.__init__(self, parent, name, size=entry["size"].value * 8)
        self.entry = entry
        res_type = self.getResType()
        if res_type in RESOURCE_TYPE:
            self._name, description, self._parser = RESOURCE_TYPE[res_type]
        else:
            self._parser = None

    def getResID(self):
        return self.entry.inode["offset"].value

    def getResType(self):
        return self.entry.inode.res_type

    def createFields(self):
        if self._parser:
            yield from self._parser(self)
        else:
            yield RawBytes(self, "content", self.size // 8)

    def createDescription(self):
        return "Resource #%u content: type=%s" % (
            self.getResID(), self.getResType())


class Header(FieldSet):
    static_size = 16 * 8

    def createFields(self):
        yield NullBytes(self, "options", 4)
        yield TimestampUnix32(self, "creation_date")
        yield UInt16(self, "maj_ver", "Major version")
        yield UInt16(self, "min_ver", "Minor version")
        yield UInt16(self, "nb_name", "Number of named entries")
        yield UInt16(self, "nb_index", "Number of indexed entries")

    def createDescription(self):
        text = "Resource header"
        info = []
        if self["nb_name"].value:
            info.append("%u name" % self["nb_name"].value)
        if self["nb_index"].value:
            info.append("%u index" % self["nb_index"].value)
        if self["creation_date"].value:
            info.append(self["creation_date"].display)
        if info:
            return "%s: %s" % (text, ", ".join(info))
        else:
            return text


class WidePascalString16(String):

    def __init__(self, parent, name, description=None,
                 strip=None, nbytes=None, truncate=None):
        Bytes.__init__(self, parent, name, 1, description)

        self._format = "WidePascalString16"
        self._strip = strip
        self._truncate = truncate
        self._character_size = 2
        self._charset = "UTF-16-LE"
        self._content_offset = 2
        self._content_size = self._character_size * self._parent.stream.readBits(
            self.absolute_address, self._content_offset * 8, self._parent.endian)
        self._size = (self._content_size + self.content_offset) * 8


class Directory(FieldSet):

    def __init__(self, parent, name, res_type=None):
        FieldSet.__init__(self, parent, name)
        nb_entries = self["header/nb_name"].value + \
            self["header/nb_index"].value
        self._size = Header.static_size + nb_entries * 64
        self.res_type = res_type

    def createFields(self):
        yield Header(self, "header")

        if MAX_NAME_PER_HEADER < self["header/nb_name"].value:
            raise ParserError("EXE resource: invalid number of name (%s)"
                              % self["header/nb_name"].value)
        if MAX_INDEX_PER_HEADER < self["header/nb_index"].value:
            raise ParserError("EXE resource: invalid number of index (%s)"
                              % self["header/nb_index"].value)

        hdr = self["header"]
        for index in range(hdr["nb_name"].value):
            yield NameOffset(self, "name[]")
        for index in range(hdr["nb_index"].value):
            yield IndexOffset(self, "index[]", self.res_type)

    def createDescription(self):
        return self["header"].description


class PE_Resource(SeekableFieldSet):

    def __init__(self, parent, name, section, size):
        SeekableFieldSet.__init__(self, parent, name, size=size)
        self.section = section

    def parseSub(self, directory, name, depth):
        indexes = []
        for index in directory.array("name"):
            if index["is_subdir"].value:
                indexes.append(index)
            self.seekByte(index["name_offset"].value)
            field = WidePascalString16(self, name.replace("directory", "name"))
            index.name_field = field
            yield field

        for index in directory.array("index"):
            if index["is_subdir"].value:
                indexes.append(index)

        # indexes.sort(key=lambda index: index["offset"].value)
        for index in indexes:
            self.seekByte(index["offset"].value)
            if depth == 1:
                res_type = index.getResType()
            else:
                res_type = directory.res_type
            yield Directory(self, name, res_type)

    def createFields(self):
        # Parse directories
        depth = 0
        subdir = Directory(self, "root")
        yield subdir
        subdirs = [subdir]
        alldirs = [subdir]
        while subdirs:
            depth += 1
            if MAX_DEPTH < depth:
                self.error(
                    "EXE resource: depth too high (%s), stop parsing directories" % depth)
                break
            newsubdirs = []
            for index, subdir in enumerate(subdirs):
                name = "directory[%u][%u][]" % (depth, index)
                try:
                    for field in self.parseSub(subdir, name, depth):
                        if field.__class__ == Directory:
                            newsubdirs.append(field)
                        yield field
                except Exception as err:
                    self.error("Unable to create directory %s: %s" %
                               (name, err))
            subdirs = newsubdirs
            alldirs.extend(subdirs)

        # Create resource list
        resources = []
        for directory in alldirs:
            for index in directory.array("index"):
                if not index["is_subdir"].value:
                    resources.append(index)

        # Parse entries
        entries = []
        for resource in resources:
            offset = resource["offset"].value
            if offset is None:
                continue
            self.seekByte(offset)
            entry = Entry(self, "entry[]", inode=resource)
            yield entry
            entries.append(entry)
        entries.sort(key=lambda entry: entry["rva"].value)

        # Parse resource content
        for entry in entries:
            try:
                offset = self.section.rva2file(entry["rva"].value)
                padding = self.seekByte(offset, relative=False)
                if padding:
                    yield padding
                yield ResourceContent(self, "content[]", entry)
            except Exception as err:
                self.warning("Error when parsing entry %s: %s" %
                             (entry.path, err))

        size = (self.size - self.current_size) // 8
        if size:
            yield PaddingBytes(self, "padding_end", size)


class NE_VersionInfoNode(FieldSet):
    TYPE_STRING = 1
    TYPE_NAME = {
        0: "binary",
        1: "string",
    }

    def __init__(self, parent, name):
        FieldSet.__init__(self, parent, name)
        self._size = alignValue(self["size"].value, 4) * 8

    def createFields(self):
        yield UInt16(self, "size", "Node size (in bytes)")
        yield UInt16(self, "data_size")
        yield CString(self, "name", charset="ISO-8859-1")

        size = paddingSize(self.current_size // 8, 4)
        if size:
            yield NullBytes(self, "padding[]", size)
        size = self["data_size"].value
        if size:
            if self["name"].value == "VS_VERSION_INFO":
                yield VersionInfoBinary(self, "value", size=size * 8)
            else:
                yield String(self, "value", size, charset="ISO-8859-1")
        while 12 <= (self.size - self.current_size) // 8:
            yield NE_VersionInfoNode(self, "node[]")
        size = (self.size - self.current_size) // 8
        if size:
            yield NullBytes(self, "padding[]", size)

    def createDescription(self):
        text = "Version info node: %s" % self["name"].value
#        if self["type"].value == self.TYPE_STRING and "value" in self:
#            text += "=%s" % self["value"].value
        return text



"""
Modplug metadata inserted into module files.

Doc:
- http://modplug.svn.sourceforge.net/viewvc/modplug/trunk/modplug/soundlib/

Author: Christophe GISQUET <christophe.gisquet@free.fr>
Creation: 10th February 2007
"""

from hachoir.field import (FieldSet,
                           UInt32, UInt16, UInt8, Int8, Float32,
                           RawBytes, String, GenericVector, ParserError)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal

MAX_ENVPOINTS = 32


def parseComments(parser):
    size = parser["block_size"].value
    if size > 0:
        yield String(parser, "comment", size)


class MidiOut(FieldSet):
    static_size = 9 * 32 * 8

    def createFields(self):
        for name in ("start", "stop", "tick", "noteon", "noteoff",
                     "volume", "pan", "banksel", "program"):
            yield String(self, name, 32, strip='\0')


class Command(FieldSet):
    static_size = 32 * 8

    def createFields(self):
        start = self.absolute_address
        size = self.stream.searchBytesLength(b"\0", False, start)
        if size > 0:
            self.info("Command: %s" % self.stream.readBytes(start, size))
            yield String(self, "command", size, strip='\0')
        yield RawBytes(self, "parameter", (self._size // 8) - size)


class MidiSFXExt(FieldSet):
    static_size = 16 * 32 * 8

    def createFields(self):
        for index in range(16):
            yield Command(self, "command[]")


class MidiZXXExt(FieldSet):
    static_size = 128 * 32 * 8

    def createFields(self):
        for index in range(128):
            yield Command(self, "command[]")


def parseMidiConfig(parser):
    yield MidiOut(parser, "midi_out")
    yield MidiSFXExt(parser, "sfx_ext")
    yield MidiZXXExt(parser, "zxx_ext")


def parseChannelSettings(parser):
    size = parser["block_size"].value // 4
    if size > 0:
        yield GenericVector(parser, "settings", size, UInt32, "mix_plugin")


def parseEQBands(parser):
    size = parser["block_size"].value // 4
    if size > 0:
        yield GenericVector(parser, "gains", size, UInt32, "band")


class SoundMixPluginInfo(FieldSet):
    static_size = 128 * 8

    def createFields(self):
        yield textHandler(UInt32(self, "plugin_id1"), hexadecimal)
        yield textHandler(UInt32(self, "plugin_id2"), hexadecimal)
        yield UInt32(self, "input_routing")
        yield UInt32(self, "output_routing")
        yield GenericVector(self, "routing_info", 4, UInt32, "reserved")
        yield String(self, "name", 32, strip='\0')
        yield String(self, "dll_name", 64, desc="Original DLL name", strip='\0')


class ExtraData(FieldSet):

    def __init__(self, parent, name, desc=None):
        FieldSet.__init__(self, parent, name, desc)
        self._size = (4 + self["size"].value) * 8

    def createFields(self):
        yield UInt32(self, "size")
        size = self["size"].value
        if size:
            yield RawBytes(self, "data", size)


class XPlugData(FieldSet):

    def __init__(self, parent, name, desc=None):
        FieldSet.__init__(self, parent, name, desc)
        self._size = (4 + self["size"].value) * 8

    def createFields(self):
        yield UInt32(self, "size")
        while not self.eof:
            yield UInt32(self, "marker")
            if self["marker"].value == 'DWRT':
                yield Float32(self, "dry_ratio")
            elif self["marker"].value == 'PORG':
                yield UInt32(self, "default_program")


def parsePlugin(parser):
    yield SoundMixPluginInfo(parser, "info")

    # Check if VST setchunk present
    size = parser.stream.readBits(
        parser.absolute_address + parser.current_size, 32, LITTLE_ENDIAN)
    if 0 < size < parser.current_size + parser._size:
        yield ExtraData(parser, "extra_data")

    # Check if XPlugData is present
    size = parser.stream.readBits(
        parser.absolute_address + parser.current_size, 32, LITTLE_ENDIAN)
    if 0 < size < parser.current_size + parser._size:
        yield XPlugData(parser, "xplug_data")


# Format: "XXXX": (type, count, name)
EXTENSIONS = {
    # WriteInstrumentHeaderStruct@Sndfile.cpp
    "XTPM": {
        "..Fd": (UInt32, 1, "Flags"),
        "..OF": (UInt32, 1, "Fade out"),
        "..VG": (UInt32, 1, "Global Volume"),
        "...P": (UInt32, 1, "Panning"),
        "..EV": (UInt32, 1, "Volume Envelope"),
        "..EP": (UInt32, 1, "Panning Envelope"),
        ".EiP": (UInt32, 1, "Pitch Envelope"),
        ".SLV": (UInt8, 1, "Volume Loop Start"),
        ".ELV": (UInt8, 1, "Volume Loop End"),
        ".BSV": (UInt8, 1, "Volume Sustain Begin"),
        ".ESV": (UInt8, 1, "Volume Sustain End"),
        ".SLP": (UInt8, 1, "Panning Loop Start"),
        ".ELP": (UInt8, 1, "Panning Loop End"),
        ".BSP": (UInt8, 1, "Panning Substain Begin"),
        ".ESP": (UInt8, 1, "Padding Substain End"),
        "SLiP": (UInt8, 1, "Pitch Loop Start"),
        "ELiP": (UInt8, 1, "Pitch Loop End"),
        "BSiP": (UInt8, 1, "Pitch Substain Begin"),
        "ESiP": (UInt8, 1, "Pitch Substain End"),
        ".ANN": (UInt8, 1, "NNA"),
        ".TCD": (UInt8, 1, "DCT"),
        ".AND": (UInt8, 1, "DNA"),
        "..SP": (UInt8, 1, "Panning Swing"),
        "..SV": (UInt8, 1, "Volume Swing"),
        ".CFI": (UInt8, 1, "IFC"),
        ".RFI": (UInt8, 1, "IFR"),
        "..BM": (UInt32, 1, "Midi Bank"),
        "..PM": (UInt8, 1, "Midi Program"),
        "..CM": (UInt8, 1, "Midi Channel"),
        ".KDM": (UInt8, 1, "Midi Drum Key"),
        ".SPP": (Int8, 1, "PPS"),
        ".CPP": (UInt8, 1, "PPC"),
        ".[PV": (UInt32, MAX_ENVPOINTS, "Volume Points"),
        ".[PP": (UInt32, MAX_ENVPOINTS, "Panning Points"),
        "[PiP": (UInt32, MAX_ENVPOINTS, "Pitch Points"),
        ".[EV": (UInt8, MAX_ENVPOINTS, "Volume Enveloppe"),
        ".[EP": (UInt8, MAX_ENVPOINTS, "Panning Enveloppe"),
        "[EiP": (UInt8, MAX_ENVPOINTS, "Pitch Enveloppe"),
        ".[MN": (UInt8, 128, "Note Mapping"),
        "..[K": (UInt32, 128, "Keyboard"),
        "..[n": (String, 32, "Name"),
        ".[nf": (String, 12, "Filename"),
        ".PiM": (UInt8, 1, "MixPlug"),
        "..RV": (UInt16, 1, "Volume Ramping"),
        "...R": (UInt16, 1, "Resampling"),
        "..SC": (UInt8, 1, "Cut Swing"),
        "..SR": (UInt8, 1, "Res Swing"),
        "..MF": (UInt8, 1, "Filter Mode"),
    },

    # See after "CODE tag dictionary", same place, elements with [EXT]
    "STPM": {
        "...C": (UInt32, 1, "Channels"),
        ".VWC": (None, 0, "CreatedWith version"),
        ".VGD": (None, 0, "Default global volume"),
        "..TD": (None, 0, "Default tempo"),
        "HIBE": (None, 0, "Embedded instrument header"),
        "VWSL": (None, 0, "LastSavedWith version"),
        ".MMP": (None, 0, "Plugin Mix mode"),
        ".BPR": (None, 0, "Rows per beat"),
        ".MPR": (None, 0, "Rows per measure"),
        "@PES": (None, 0, "Chunk separator"),
        ".APS": (None, 0, "Song Pre-amplification"),
        "..MT": (None, 0, "Tempo mode"),
        "VTSV": (None, 0, "VSTi volume"),
    }
}


class MPField(FieldSet):

    def __init__(self, parent, name, ext, desc=None):
        FieldSet.__init__(self, parent, name, desc)
        self.ext = ext
        self.info(self.createDescription())
        self._size = (6 + self["data_size"].value) * 8

    def createFields(self):
        # Identify tag
        code = self.stream.readBytes(self.absolute_address, 4)
        if code in self.ext:
            cls, count, comment = self.ext[code]
        else:
            cls, count, comment = RawBytes, 1, "Unknown tag"

        # Header
        yield String(self, "code", 4, comment)
        yield UInt16(self, "data_size")

        # Data
        if not cls:
            size = self["data_size"].value
            if size > 0:
                yield RawBytes(self, "data", size)
        elif cls in (String, RawBytes):
            yield cls(self, "value", count)
        else:
            if count > 1:
                yield GenericVector(self, "values", count, cls, "item")
            else:
                yield cls(self, "value")

    def createDescription(self):
        return "Element '%s', size %i" % \
               (self["code"]._description, self["data_size"].value)


def parseFields(parser):
    # Determine field names
    ext = EXTENSIONS[parser["block_type"].value]
    if ext is None:
        raise ParserError("Unknown parent '%s'" % parser["block_type"].value)

    # Parse fields
    addr = parser.absolute_address + parser.current_size
    while not parser.eof and parser.stream.readBytes(addr, 4) in ext:
        field = MPField(parser, "field[]", ext)
        yield field
        addr += field._size

    # Abort on unknown codes
    parser.info("End of extension '%s' when finding '%s'" %
                (parser["block_type"].value, parser.stream.readBytes(addr, 4)))


class ModplugBlock(FieldSet):
    BLOCK_INFO = {
        "TEXT": ("comment", True, "Comment", parseComments),
        "MIDI": ("midi_config", True, "Midi configuration", parseMidiConfig),
        "XFHC": ("channel_settings", True, "Channel settings", parseChannelSettings),
        "XTPM": ("instrument_ext", False, "Instrument extensions", parseFields),
        "STPM": ("song_ext", False, "Song extensions", parseFields),
    }

    def __init__(self, parent, name, desc=None):
        FieldSet.__init__(self, parent, name, desc)
        self.parseBlock = parsePlugin

        t = self["block_type"].value
        self.has_size = False
        if t in self.BLOCK_INFO:
            self._name, self.has_size, desc, parseBlock = self.BLOCK_INFO[t]
            if callable(desc):
                self.createDescription = lambda: desc(self)
            if parseBlock:
                self.parseBlock = lambda: parseBlock(self)

        if self.has_size:
            self._size = 8 * (self["block_size"].value + 8)

    def createFields(self):
        yield String(self, "block_type", 4)
        if self.has_size:
            yield UInt32(self, "block_size")

        if self.parseBlock:
            yield from self.parseBlock()

        if self.has_size:
            size = self["block_size"].value - (self.current_size // 8)
            if size > 0:
                yield RawBytes(self, "data", size, "Unknown data")


def ParseModplugMetadata(parser):
    while not parser.eof:
        block = ModplugBlock(parser, "block[]")
        yield block
        if block["block_type"].value == "STPM":
            break

    # More undocumented stuff: date ?
    size = (parser._size - parser.absolute_address - parser.current_size) // 8
    if size > 0:
        yield RawBytes(parser, "info", size)



from hachoir.parser.file_system.ext2 import EXT2_FS  # noqa
from hachoir.parser.file_system.fat import FAT12, FAT16, FAT32  # noqa
from hachoir.parser.file_system.mbr import MSDos_HardDrive  # noqa
from hachoir.parser.file_system.ntfs import NTFS  # noqa
from hachoir.parser.file_system.iso9660 import ISO9660  # noqa
from hachoir.parser.file_system.reiser_fs import REISER_FS  # noqa
from hachoir.parser.file_system.linux_swap import LinuxSwapFile  # noqa



from datetime import date, datetime
import re
from hachoir.core.language import Language
from locale import setlocale, LC_ALL
from time import strptime
from hachoir.metadata.timezone import createTimezone
from hachoir.metadata import config

NORMALIZE_REGEX = re.compile("[-/.: ]+")
YEAR_REGEX1 = re.compile("^([0-9]{4})$")

# Date regex: YYYY-MM-DD (US format)
DATE_REGEX1 = re.compile("^([0-9]{4})~([01][0-9])~([0-9]{2})$")

# Date regex: YYYY-MM-DD HH:MM:SS (US format)
DATETIME_REGEX1 = re.compile("^([0-9]{4})~([01][0-9])~([0-9]{2})"
                             "~([0-9]{1,2})~([0-9]{2})~([0-9]{1,2})$")

# Datetime regex: "MM-DD-YYYY HH:MM:SS" (FR format)
DATETIME_REGEX2 = re.compile("^([01]?[0-9])~([0-9]{2})~([0-9]{4})"
                             "~([0-9]{1,2})~([0-9]{2})~([0-9]{1,2})$")

# Timezone regex: "(...) +0200"
TIMEZONE_REGEX = re.compile("^(.*)~([+-][0-9]{2})00$")

# Timestmap: 'February 2007'
MONTH_YEAR = "%B~%Y"

# Timestmap: 'Sun Feb 24 15:51:09 2008'
RIFF_TIMESTAMP = "%a~%b~%d~%H~%M~%S~%Y"

# Timestmap: 'Thu, 19 Jul 2007 09:03:57'
ISO_TIMESTAMP = "%a,~%d~%b~%Y~%H~%M~%S"


def parseDatetime(value):
    """
    Year and date:
    >>> parseDatetime("2000")
    (datetime.date(2000, 1, 1), '2000')
    >>> parseDatetime("2004-01-02")
    datetime.date(2004, 1, 2)

    Timestamp:
    >>> parseDatetime("2004-01-02 18:10:45")
    datetime.datetime(2004, 1, 2, 18, 10, 45)
    >>> parseDatetime("2004-01-02 18:10:45")
    datetime.datetime(2004, 1, 2, 18, 10, 45)

    Timestamp with timezone:
    >>> parseDatetime('Thu, 19 Jul 2007 09:03:57 +0000')
    datetime.datetime(2007, 7, 19, 9, 3, 57, tzinfo=<TimezoneUTC delta=0, name='UTC'>)
    >>> parseDatetime('Thu, 19 Jul 2007 09:03:57 +0200')
    datetime.datetime(2007, 7, 19, 9, 3, 57, tzinfo=<Timezone delta=2:00:00, name='+0200'>)
    """
    value = NORMALIZE_REGEX.sub("~", value.strip())
    regs = YEAR_REGEX1.match(value)
    if regs:
        try:
            year = int(regs.group(1))
            return (date(year, 1, 1), str(year))
        except ValueError:
            pass
    regs = DATE_REGEX1.match(value)
    if regs:
        try:
            year = int(regs.group(1))
            month = int(regs.group(2))
            day = int(regs.group(3))
            return date(year, month, day)
        except ValueError:
            pass
    regs = DATETIME_REGEX1.match(value)
    if regs:
        try:
            year = int(regs.group(1))
            month = int(regs.group(2))
            day = int(regs.group(3))
            hour = int(regs.group(4))
            min = int(regs.group(5))
            sec = int(regs.group(6))
            return datetime(year, month, day, hour, min, sec)
        except ValueError:
            pass
    regs = DATETIME_REGEX2.match(value)
    if regs:
        try:
            month = int(regs.group(1))
            day = int(regs.group(2))
            year = int(regs.group(3))
            hour = int(regs.group(4))
            min = int(regs.group(5))
            sec = int(regs.group(6))
            return datetime(year, month, day, hour, min, sec)
        except ValueError:
            pass
    current_locale = setlocale(LC_ALL, "C")
    try:
        match = TIMEZONE_REGEX.match(value)
        if match:
            without_timezone = match.group(1)
            delta = int(match.group(2))
            delta = createTimezone(delta)
        else:
            without_timezone = value
            delta = None
        try:
            timestamp = strptime(without_timezone, ISO_TIMESTAMP)
            arguments = list(timestamp[0:6]) + [0, delta]
            return datetime(*arguments)
        except ValueError:
            pass

        try:
            timestamp = strptime(without_timezone, RIFF_TIMESTAMP)
            arguments = list(timestamp[0:6]) + [0, delta]
            return datetime(*arguments)
        except ValueError:
            pass

        try:
            timestamp = strptime(value, MONTH_YEAR)
            arguments = list(timestamp[0:3])
            return date(*arguments)
        except ValueError:
            pass
    finally:
        setlocale(LC_ALL, current_locale)
    return None


def setDatetime(meta, key, value):
    if isinstance(value, str):
        return parseDatetime(value)
    elif isinstance(value, (date, datetime)):
        return value
    return None


def setLanguage(meta, key, value):
    """
    >>> setLanguage(None, None, "fre")
    <Language 'French', code='fre'>
    >>> setLanguage(None, None, u"ger")
    <Language 'German', code='ger'>
    """
    return Language(value)


def setTrackTotal(meta, key, total):
    """
    >>> setTrackTotal(None, None, "10")
    10
    """
    try:
        return int(total)
    except ValueError:
        meta.warning("Invalid track total: %r" % total)
        return None


def setTrackNumber(meta, key, number):
    if isinstance(number, int):
        return number
    if "/" in number:
        number, total = number.split("/", 1)
        meta.track_total = total
    try:
        return int(number)
    except ValueError:
        meta.warning("Invalid track number: %r" % number)
        return None


def normalizeString(text):
    if config.RAW_OUTPUT:
        return text
    return text.strip(" \t\v\n\r\0")



from hachoir.parser import QueryParser
from hachoir.regex import PatternMatching


# XXX hachoir.regex uses str but the rest of hachoir uses bytes,
# which means we have to convert bytes to str using latin1 encoding
# (the closest "raw bytes" encoding) in order for matching to work.
class HachoirPatternMatching(PatternMatching):

    def __init__(self, categories=None, parser_ids=None):
        PatternMatching.__init__(self)

        # Load parser list
        tags = []
        if categories:
            tags += [("category", cat) for cat in categories]
        if parser_ids:
            tags += [("id", parser_id) for parser_id in parser_ids]
        if tags:
            tags += [None]
        parser_list = QueryParser(tags)

        # Create string patterns
        for parser in parser_list:
            for (magic, offset) in parser.getParserTags().get("magic", ()):
                self.addString(magic.decode('latin1'), (offset, parser))

        # Create regex patterns
        for parser in parser_list:
            for (regex, offset) in parser.getParserTags().get("magic_regex", ()):
                self.addRegex(regex.decode('latin1'), (offset, parser))
        self.commit()

    def search(self, data):
        for start, stop, item in PatternMatching.search(self, data.decode('latin1')):
            yield (item.user[1], start * 8 - item.user[0])



"""
GNU ar archive : archive file (.a) and Debian (.deb) archive.
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, ParserError,
                           String, RawBytes, UnixLine)
from hachoir.core.endian import BIG_ENDIAN


class ArchiveFileEntry(FieldSet):

    def createFields(self):
        yield UnixLine(self, "header", "Header")
        info = self["header"].value
        info = info.split()
        if len(info) < 3:
            raise ParserError("Invalid file entry header: %r" % info)
        size = int(info[-2])
        if 0 < size:
            yield RawBytes(self, "content", size, "File data")

    def createDescription(self):
        return "File entry (%s)" % self["header"].value.split()[0]


class ArchiveFile(Parser):
    endian = BIG_ENDIAN
    MAGIC = b'!<arch>\n'
    PARSER_TAGS = {
        "id": "unix_archive",
        "category": "archive",
        "file_ext": ("a", "deb"),
        "mime":
            ("application/x-debian-package",
             "application/x-archive",
             "application/x-dpkg"),
        "min_size": (8 + 13) * 8,  # file signature + smallest file as possible
        "magic": ((MAGIC, 0),),
        "description": "Unix archive"
    }

    def validate(self):
        if self.stream.readBytes(0, len(self.MAGIC)) != self.MAGIC:
            return "Invalid magic string"
        return True

    def createFields(self):
        yield String(self, "id", 8, "Unix archive identifier (\"<!arch>\")", charset="ASCII")
        while not self.eof:
            data = self.stream.readBytes(self.current_size, 1)
            if data == b"\n":
                yield RawBytes(self, "empty_line[]", 1, "Empty line")
            else:
                yield ArchiveFileEntry(self, "file[]", "File")



from hachoir.parser import createParser
from hachoir.core.tools import makePrintable
from hachoir.metadata import extractMetadata
from hachoir.core.i18n import initLocale
from sys import argv, stderr, exit
from os import walk
from os.path import join as path_join
from fnmatch import fnmatch
import codecs

OUTPUT_FILENAME = "metadata.csv"


class Extractor:

    def __init__(self, directory, fields):
        self.directory = directory
        self.fields = fields
        self.charset = "UTF-8"
        self.total = 0
        self.invalid = 0

    def main(self):
        output = codecs.open(OUTPUT_FILENAME, "w", self.charset)
        for filename in self.findFiles(self.directory, '*.doc'):
            self.total += 1
            line = self.processFile(filename)
            if line:
                print(line, file=output)
            else:
                self.invalid += 1
        output.close()
        self.summary()

    def summary(self):
        print(file=stderr)
        print("Valid files: %s" % (self.total - self.invalid), file=stderr)
        print("Invalid files: %s" % self.invalid, file=stderr)
        print("Total files: %s" % self.total, file=stderr)
        print(file=stderr)
        print("Result written into %s" % OUTPUT_FILENAME, file=stderr)

    def findFiles(self, directory, pattern):
        for dirpath, dirnames, filenames in walk(directory):
            for filename in filenames:
                if not fnmatch(filename.lower(), pattern):
                    continue
                yield path_join(dirpath, filename)

    def processFile(self, filename):
        print("[%s] Process file %s..." % (self.total, filename))
        parser = createParser(filename)
        if not parser:
            print("Unable to parse file", file=stderr)
            return None
        try:
            metadata = extractMetadata(parser)
        except Exception as err:
            print("Metadata extraction error: %s" % str(err), file=stderr)
            return None
        if not metadata:
            print("Unable to extract metadata", file=stderr)
            return None

        filename = makePrintable(filename, self.charset)
        line = [filename]
        for field in self.fields:
            value = metadata.getText(field, '')
            value = makePrintable(value, self.charset)
            line.append(value)
        return '; '.join(line)


def main():
    initLocale()
    if len(argv) != 3:
        print("usage: %s directory fields" % argv[0], file=stderr)
        print(file=stderr)
        print("eg. %s . title,creation_date" % argv[0], file=stderr)
        exit(1)
    directory = argv[1]
    fields = [field.strip() for field in argv[2].split(",")]
    Extractor(directory, fields).main()
